#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# arc class
#

import warnings
import math

import circle
import style
import util
import graphicobject
import linetype
import color
import point
import tolerance
import maptree
import quadtree

_dtr = math.pi/180.0
_rtd = 180.0/math.pi

class Arc(graphicobject.new_GraphicObject):
    """A class for Arcs.

An Arc has four attributes:

center: A Point object
radius: The Arc's radius
start_angle:
end_angle:

An Arc has the following methods:

{get/set}Center(): Get/Set the center Point of an Arc.
{get/set}Radius(): Get/Set the radius of an Arc.
{get/set}StartAngle(): Get/Set the starting angle of an Arc.
{get/set}EndAngle(): Get/Set the end angle of an arc
move(): Move the Arc.
length(): Return the length of an Arc.
area(): Return the area of an Arc.
mapPoint(): Find the nearest Point on the Arc to some other Point.
mapCoords(): Find the nearest Point on the Arc to a coordinate pair.
inRegion(): Returns whether or not a Arc can be seen in a bounded area.
clone(): Return an indentical copy of a Arc.
    """

    __defstyle = style.Style(u'Default Arc Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    messages = {
        'moved' : True,
        'center_changed' : True,
        'radius_changed' : True,
        'start_angle_changed' : True,
        'end_angle_changed' : True
        }

    def __init__(self, center, radius, start_angle, end_angle,
                 st=None, lt=None, col=None, th=None):
        """Initialize a Arc.

Arc(center, radius, start_angle, end_angle)

The center should be a Point, or a two-entry tuple of floats,
and the radius should be a float greater than 0.
        """
        _cp = center
        if not isinstance(_cp, point.Point):
            _cp = point.Point(center)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _st = st
        _sa = util.make_c_angle(start_angle)
        _ea = util.make_c_angle(end_angle)
        if _st is None:
            _st = Arc.__defstyle
        graphicobject.new_GraphicObject.__init__(self, _st, lt, col, th)
        self.__radius = _r
        self.__sa = _sa
        self.__ea = _ea
        self.__center = _cp
        _cp.connect('moved', self, Arc.movePoint)
        _cp.storeUser(self)

    def __eq__(self, obj):
        """Compare a Arc to another for equality.
        """
        if not isinstance(obj, Arc):
            return False
        _val = False
        if obj is self:
            _val = True
        else:
            if ((self.__center == obj.getCenter()) and
                (abs(self.__radius - obj.getRadius()) < 1e-10) and
                (abs(self.__sa - obj.getStartAngle()) < 1e-10) and
                (abs(self.__ea - obj.getEndAngle()) < 1e-10)):
                    _val = True
        return _val

    def __ne__(self, obj):
        """Compare a Arc to another for inequality.
        """
        if not isinstance(obj, Arc):
            return True
        _val = True
        if obj is self:
            _val = False
        else:
            if ((self.__center == obj.getCenter()) and
                (abs(self.__radius - obj.getRadius()) < 1e-10) and
                (abs(self.__sa - obj.getStartAngle()) < 1e-10) and
                (abs(self.__ea - obj.getEndAngle()) < 1e-10)):
                    _val = False
        return _val

    def getCenter(self):
        """Return the center Point of the Arc.

getCenter()
        """
        return self.__center

    def setCenter(self, c):
        """Set the center Point of the Arc.

setCenter(c)

The argument must be a Point or a tuple containing
two float values.
        """
        if self.isLocked():
            raise RuntimeError, "Setting center not allowed - object locked."
        _cp = self.__center
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid center point: " + `c`
        if _cp is not c:
            _cp.disconnect(self)
            _cp.freeUser(self)
            self.__center = c
            self.sendMessage('center_changed', _cp)
            c.connect('moved', self, Arc.movePoint)
            c.storeUser(self)
            if abs(_cp.x - c.x) > 1e-10 or abs(_cp.y - c.y) > 1e-10:
                self.sendMessage('moved', _cp.x, _cp.y, self.__radius,
                                 self.__sa, self.__ea)
            self.modified()

    center = property(getCenter, setCenter, None, "Arc center")

    def getRadius(self):
        """Return the radius of the the Arc.

getRadius()
        """
        return self.__radius

    def setRadius(self, radius):
        """Set the radius of the Arc.

setRadius(radius)

The argument must be float value greater than 0.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _cr = self.__radius
        if abs(_cr - _r) > 1e-10:
            self.__radius = _r
            self.sendMessage('radius_changed', _cr)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, _cr, self.__sa, self.__ea)
            self.modified()

    radius = property(getRadius, setRadius, None, "Arc radius")

    def getStartAngle(self):
        """Return the start_angle for the Arc.

getStartAngle()
        """
        return self.__sa

    def setStartAngle(self, angle):
        """Set the start_angle for the Arc.

setStartAngle(angle)

The argument angle should be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Setting start angle not allowed - object locked."
        _sa = self.__sa
        _angle = util.make_c_angle(angle)
        if abs(_sa - _angle) > 1e-10:
            self.__sa = _angle
            self.sendMessage('start_angle_changed', _sa)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, self.__radius, _sa, self.__ea)
            self.modified()

    start_angle = property(getStartAngle, setStartAngle, None,
                           "Start angle for the Arc.")

    def getEndAngle(self):
        """Return the end_angle for the Arc.

getEndAngle()
        """
        return self.__ea

    def setEndAngle(self, angle):
        """Set the end_angle for the Arc.

setEndAngle(angle)

The argument angle should be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Setting end angle not allowed - object locked."
        _ea = self.__ea
        _angle = util.make_c_angle(angle)
        if abs(_ea - _angle) > 1e-10:
            self.__ea = _angle
            self.sendMessage('end_angle_changed', _ea)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, self.__radius, self.__sa, _ea)
            self.modified()

    end_angle = property(getEndAngle, setEndAngle, None,
                         "End angle for the Arc.")

    def getAngle(self):
        """Return the angular sweep of the Arc.

getAngle()
        """
        _sa = self.__sa
        _ea = self.__ea
        if abs(_ea - _sa) < 1e-10:
            _angle = 360.0
        elif _ea > _sa:
            _angle = _ea - _sa
        else:
            _angle = 360.0 - _sa + _ea
        return _angle

    def throughAngle(self, angle):
        """Return True if an arc passes through some angle

throughAngle(angle)

The argument angle should be a float value. This method returns
True if the arc exists at that angle, otherwise the method returns False.
        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        _angle = math.fmod(_angle, 360.0)
        if _angle < 0.0:
            _angle = _angle + 360.0
        _sa = self.__sa
        _ea = self.__ea
        _val = True
        if abs(_sa - _ea) > 1e-10:
            if _sa > _ea:
                if _angle > _ea and _angle < _sa:
                    _val = False
            else:
                if _angle > _ea or _angle < _sa:
                    _val = False
        return _val

    def getEndpoints(self):
        """Return where the two endpoints for the arc-segment lie.

getEndpoints(self)

This function returns two tuples, each containing the x-y coordinates
of the arc endpoints. The first tuple corresponds to the endpoint at
the start_angle, the second to the endpoint at the end_angle.
        """
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _sa = self.__sa
        _sax = _cx + _r * math.cos(_sa * _dtr)
        _say = _cy + _r * math.sin(_sa * _dtr)
        _ea = self.__ea
        _eax = _cx + _r * math.cos(_ea * _dtr)
        _eay = _cy + _r * math.sin(_ea * _dtr)
        return (_sax, _say), (_eax, _eay)

    def length(self):
        """Return the length of the Arc.

length()
        """
        return 2.0 * math.pi * self.__radius * (self.getAngle()/360.0)

    def area(self):
        """Return the area enclosed by the Arc.

area()
        """
        return math.pi * pow(self.__radius, 2) * (self.getAngle()/360.0)

    def move(self, dx, dy):
        """Move a Arc.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__center.getCoords()
            self.mute()
            self.__center.move(_dx, _dy)
            self.unmute()
            self.sendMessage('moved', _x, _y, self.__radius,
                             self.__sa, self.__ea)
            self.modified()


    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Arc to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the x-coordinate
y: A Float value giving the y-coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the Arc. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _dist = math.hypot((_x - _cx), (_y - _cy))
        if abs(_dist - _r) < _t:
            _angle = _rtd * math.atan2((_y - _cy),(_x - _cx))
            if _angle < 0.0:
                _angle = _angle + 360.0
            if self.throughAngle(_angle):
                _xoff = _r * math.cos(_angle)
                _yoff = _r * math.sin(_angle)
                return (_cx + _xoff), (_cy + _yoff)
        return None

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Arc to the Point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by Point object to an
actual Point on the Arc. If the distance between the actual
Point and the Point used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        warnings.warn("mapPoint() is deprecated - use mapCoords()",
        DeprecationWarning, stacklevel=2)
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an Arc exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument fully indicates whether or not the Arc
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        _val = False
        _bits = 0
        #
        # calculate distances from center to region boundary
        #
        if abs(_xc - _xmin) < _r: _bits = _bits | 1 # left edge
        if abs(_xc - _xmax) < _r: _bits = _bits | 2 # right edge
        if abs(_yc - _ymin) < _r: _bits = _bits | 4 # bottom edge
        if abs(_yc - _ymax) < _r: _bits = _bits | 8 # top edge
        if _bits == 0:
            #
            # if the arc center is in region then the entire
            # arc is visible since the distance from the center
            # to any edge is greater than the radius. If the center
            # is not in the region then the arc is not visible in
            # the region because the distance to any edge is greater
            # than the radius, and so one of the bits should have been
            # set ...
            #
            if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                _val = True
        else:
            _val = True
            #
            # calculate distance to corners of region
            #
            if math.hypot((_xc - _xmin), (_yc - _ymax)) < _r:
                _bits = _bits | 0x10 # upper left
            if math.hypot((_xc - _xmax), (_yc - _ymin)) < _r:
                _bits = _bits | 0x20 # lower right
            if math.hypot((_xc - _xmin), (_yc - _ymin)) < _r:
                _bits = _bits | 0x40 # lower left
            if math.hypot((_xc - _xmax), (_yc - _ymax)) < _r:
                _bits = _bits | 0x80 # upper right
            #
            # if all bits are set then distance from arc center
            # to region endpoints is less than radius - arc
            # entirely outside the region
            #
            if _bits == 0xff or fully:
                _val = False
            #
            # if the test value is still true, check that the
            # arc boundary can overlaps with the region
            #
            if _val:
                _sa = self.__sa
                _ea = self.__ea
                _xs = _xc + (_r * math.cos(_sa * _dtr))
                _ys = _yc + (_r * math.sin(_sa * _dtr))
                _xe = _xc + (_r * math.cos(_ea * _dtr))
                _ye = _xc + (_r * math.sin(_ea * _dtr))
                _axmin = min(_xs, _xe)
                if self.throughAngle(180.0):
                    _axmin = _xc - _r
                if _axmin > _xmax:
                    return False
                _aymin = min(_ys, _ye)
                if self.throughAngle(270.0):
                    _aymin = _yc - _r
                if _aymin > _ymax:
                    return False
                _axmax = max(_xs, _xe)
                if self.throughAngle(0.0):
                    _axmax = _xc + _r
                if _axmax < _xmin:
                    return False
                _aymax = max(_ys, _ye)
                if self.throughAngle(90.0):
                    _aymax = _yc + _r
                if _aymax < _ymin:
                    return False
        return _val

    def getBounds(self):
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        _sa = self.__sa
        _ea = self.__ea
        _xs = _xc + (_r * math.cos(_sa * _dtr))
        _ys = _yc + (_r * math.sin(_sa * _dtr))
        _xe = _xc + (_r * math.cos(_ea * _dtr))
        _ye = _xc + (_r * math.sin(_ea * _dtr))
        _xmin = min(_xs, _xe)
        _ymin = min(_ys, _ye)
        _xmax = max(_xs, _xe)
        _ymax = max(_ys, _ye)
        if self.throughAngle(0.0):
            _xmax = _xc + _r
        if self.throughAngle(90.0):
            _ymax = _yc + _r
        if self.throughAngle(180.0):
            _xmin = _xc - _r
        if self.throughAngle(270.0):
            _ymin = _yc - _r
        return _xmin, _ymin, _xmax, _ymax

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _cp = self.__center
        if p is not _cp:
            raise ValueError, "Point is not arc center: " + `p`
        _x, _y = _cp.getCoords()
        self.sendMessage('moved', _x, _y, self.__radius, self.__sa, self.__ea)
        self.modified()

    def clone(self):
        """Create an identical copy of a Arc

clone()
        """
        _cp = self.__center.clone()
        _r = self.__radius
        _sa = self.__sa
        _ea = self.__ea
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Arc(_cp, _r, _sa, _ea, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Arc.messages:
            return True
        return graphicobject.new_GraphicObject.sendsMessage(self, m)

class old_Arc(circle.Circle):
    """A class representing Arcs.

An Arc is derived from a Circle, so it has
all the attributes and methods as that object.
An Arc has two additional attributes:

start_angle: A float (0.0 < start_angle < 360.0)
end_angle: A float (0.0 < end_angle < 360.0)

In addition to the Circle methods, an Arc has
the following methods:

{get/set}StartAngle(): Get/Set the Arc's start angle.
{get/set}EndAngle(): Get/Set the Arc's end angle.
getAngle(): Get the angular sweep of the arc.
throughAngle(): Test if an arc exists at some angle.
getEndpoints(): Get the points at the ends of the arc.
length(): Get the length of the Arc.

Arc are counter-clockwise, from the start angle to the end angle.
    """

    # member functions

    def __init__(self, center, radius, start_angle, end_angle,
                 st=style.Style('Default'), lt=None, col=None, t=None):
        """Initialize an Arc object.

Arc(center, radius, start_angle, end_angle, st [ ,lt, col t])

center: A Point object for the Arc's center
radius: The Arc's radius
start_angle: A float for the Arc's starting_angle
end_angle: A float for the Arc's end_angle
st: A style object

Optional arguments:
lt: A Linetype object that overrides the Style linetype
col: A Color object that overrides the Style color
t: A float that overrides the Style thickness

        """
        _sa = util.make_c_angle(start_angle)
        _ea = util.make_c_angle(end_angle)
        circle.Circle.__init__(self, center, radius, st, lt, col, t)
        self.__sa = _sa
        self.__ea = _ea

    def __str__(self):
        _c = self.getCenter().getCoords()
        _r = self.getRadius()
        return "Arc: Center: %s Radius: %g Start Angle: %g End Angle: %g" % (_c, _r, self.__sa, self.__ea)

    def __eq__(self, obj):
        """Compare one Arc to another for equality.
        """
        if not isinstance(obj, Arc):
            return False
        _val = False
        if circle.Circle.__eq__(self, obj):
            if abs(self.__sa - obj.getStartAngle()) < 1e-10:
                if abs(self.__ea - obj.getEndAngle()) < 1e-10:
                    _val = True
        return _val

    def __ne__(self, obj):
        """Compare one Arc to another for inequality.
        """
        if not isinstance(obj, Arc):
            return True
        _val = True
        if circle.Circle.__eq__(self, obj):
            if abs(self.__sa - obj.getStartAngle()) < 1e-10:
                if abs(self.__ea - obj.getEndAngle()) < 1e-10:
                    _val = False
        return _val

    def __cmp__(self, obj):
        """Compare two Arc objects or an Arc and a Point.

As Arcs are derived from Circles, the comparison first does
a Circle comparison on the two Arc objects. If the result of
that comparison is that the two are equal, then the Arcs are
compared based on two values:

start_angle of A1 < start_angle of A2 ==> return -1
start_angle of A1 > start_angle of A2 ==> return 1

If those are equal, the "sweep" of the Arc is calculated.

sweep of A1 < sweep of A2 ==> return -1
sweep of A1 > sweep of A2 ==> return 1

If we get here, return 0

As Arc objects are all Circle objects, the Arc/Point
comparison is done in the same manner.
        """

        if not isinstance(obj, (Arc, circle._Circle, point._Point)):
            raise TypeError, "Invalid object for Arc comparison: " + `obj`
        _res = circle.Circle.__cmp__(self, obj)
        if _res == 0 and isinstance(obj, Arc):
            _ssa = self.__sa
            _asa = obj.__sa
            if abs(_ssa - _asa) < 1e-10:
                _sa = self.getAngle()
                _aa = obj.getAngle()
                if abs(_sa - _aa) < 1e-10:
                    _res = 0
                elif _sa > _aa:
                    _res = 1
                else:
                    _res = -1
            elif _ssa > _asa:
                _res = 1
            else:
                _res = -1
        return _res

    def getStartAngle(self):
        """Return the start_angle for the Arc.

getStartAngle()
        """
        return self.__sa

    def setStartAngle(self, angle):
        """Set the start_angle for the Arc.

setStartAngle(angle)

The argument angle should be a float.
        """
        _a = util.make_c_angle(angle)
        self.__sa = _a
        self.modified()

    start_angle = property(getStartAngle, setStartAngle, None,
                           "Start angle for the Arc.")

    def getEndAngle(self):
        """Return the end_angle for the Arc.

getEndAngle()
        """
        return self.__ea

    def setEndAngle(self, angle):
        """Set the end_angle for the Arc.

setEndAngle(angle)

The argument angle should be a float.
        """
        _a = util.make_c_angle(angle)
        self.__ea = _a
        self.modified()

    end_angle = property(getEndAngle, setEndAngle, None,
                         "End angle for the Arc.")

    def getAngle(self):
        """Return the angular sweep of the Arc.

getAngle()
        """
        _sa = self.__sa
        _ea = self.__ea
        if abs(_ea - _sa) < 1e-10:
            _angle = 360.0
        elif _ea > _sa:
            _angle = _ea - _sa
        else:
            _angle = 360.0 - _sa + _ea
        return _angle

    def throughAngle(self, angle):
        """Return True if an arc passes through some angle

throughAngle(angle)

The argument "angle" should be a float value from 0.0 to 360.0.
This method returns True if the arc exists at that angle, otherwise
the method returns False
        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        if _angle < 0.0 or angle > 360.0:
            raise ValueError, "Invalid angle: %g" % _angle
        _sa = self.__sa
        _ea = self.__ea
        _val = False
        if abs(_sa - _ea) < 1e-10:
            _val = True
        elif _sa > _ea:
            if (((_sa - 1e-10) < _angle < (360.0 + 1e-10)) or
                (-1e-10 < _angle < (_ea + 1e-10))):
                _val = True
        else:
            if ((_sa - 1e-10) < _angle < (_ea + 1e-10)):
                _val = True
        return _val

    def getEndpoints(self):
        """Return where the two endpoints for the arc-segment lie.

getEndpoints(self)

This function returns two tuples, each containing the x-y coordinates
of the arc endpoints. The first tuple corresponds to the endpoint at
the start_angle, the second to the endpoint at the end_angle.
        """
        _cx, _cy = self.getCenter().getCoords()
        _r = self.getRadius()
        _sa = self.__sa
        _sax = _cx + _r * math.cos(_sa * _dtr)
        _say = _cy + _r * math.sin(_sa * _dtr)
        _ea = self.__ea
        _eax = _cx + _r * math.cos(_ea * _dtr)
        _eay = _cy + _r * math.sin(_ea * _dtr)
        return ((_sax, _say), (_eax, _eay))

    def circumference(self):
        """This method overrides the Circle::circumference() method.

circumference()

This method does nothing - use length() for an arc.
        """
        pass

    def length(self):
        """Return the length of the Arc.

length()
        """
        return 2.0 * math.pi * self.getRadius() * (self.getAngle()/360.0)

    def area(self):
        """Return the area enclosed by the Arc.

area()
        """
        return math.pi * pow(self.getRadius(), 2) * (self.getAngle()/360.0)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Arc to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the Arc. If the distance between the actual
Point and the Point used as an argument is less than the Tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _ap = None
        _pt = circle._Circle.mapPoint(self, p, tol)
        if _pt is not None:
            _xc, _yc = self.getCenter().getCoords()
            _xp, _yp = _pt.getCoords()
            _angle = _rtd * math.atan2((_yp - _yc), (_xp - _xc))
            if _angle < 0.0:
                _angle = _angle + 360.0
            if self.throughAngle(_angle):
                _ap = point.Point(_xp, _yp)
        return _ap

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an Arc exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the _Circle
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _x, _y = self.getCenter().getCoords()
        _r = self.getRadius()
        _cxmin = _x - _r
        _cymin = _y - _r
        _cxmax = _x + _r
        _cymax = _y + _r
        _val = False
        if (_xmin < _cxmin and _xmax > _cxmax and
            _ymin < _cymin and _ymax > _cymax):
            _val = True
        elif (not fully and
              not (_cxmax < _xmin or _cymax < _ymin or
                   _cxmin > _xmax or _cymin > _ymax)):
            _d1 = math.hypot((_xmin - _x), (_ymin - _y))
            _d2 = math.hypot((_xmin - _x), (_ymax - _y))
            _d3 = math.hypot((_xmax - _x), (_ymax - _y))
            _d4 = math.hypot((_xmax - _x), (_ymin - _y))
            _dmin = min(_d1, _d2, _d3, _d4)
            _dmax = max(_d1, _d2, _d3, _d4)
            _p1, _p2 = self.getEndpoints()
            _p1x, _p1y = _p1
            _p2x, _p2y = _p2
            if _xmin < _p1x < _xmax and _ymin < _p1y < _ymax:
                _val = True
            elif _xmin < _p2x < _xmax and _ymin < _p2y < _ymax:
                _val = True
            else:
                if _xmin < _x < _xmax and _ymin < _y < _ymax:
                    _dmin = -1e-10
                if _x > _xmax and _ymin < _y < _ymax:
                    _dmin = _x - _xmax
                elif _x < _xmin and _ymin < _y < _ymax:
                    _dmin = _xmin - _x
                elif _y > _ymax and _xmin < _x < _xmax:
                    _dmin = _y - _ymax
                elif _y < _ymin and _xmin < _x < _xmax:
                    _dmin = _ymin - _y
                if _dmin < _r < _dmax:
                    _da = _rtd * math.atan2((_ymin - _y), (_xmin - _x))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self.throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * math.atan2((_ymin - _y), (_xmax - _x))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self.throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * math.atan2((_ymax - _y), (_xmax - _x))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self.throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * math.atan2((_ymax - _y), (_xmin - _x))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self.throughAngle(_da)
        return _val

    def clone(self):
        """Create an identical copy of an Arc.

clone()
        """
        _cp = self.getCenter().clone()
        _r = self.getRadius()
        _sa = self.__sa
        _ea = self.__ea
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Arc(_cp, _r, _sa, _ea, _st, _lt, _col, _th)

#
# static functions for Arc class
#

    def test_angle(s, e, a):
        """Returns if an angle lies between the start and end angle of an arc.

test_angle(s, e, a)

The arguments are:

s: arc start angle
e: arc end angle
a: angle being tested
        """
        _val = False
        if ((abs(e - s) < 1e-10) or
            ((s > e) and
             ((s <= a <= 360.0) or (0.0 <= a <= e))) or
            (s <= a <= e)):
            _val = True
        return _val

    test_angle = staticmethod(test_angle)

class ArcTree(maptree.PointMapTree):
    """A Tree class specific for Arc objects.

This class is based on the PointMapTree class, and
holds only Arc objects.
    """
    def __init__(self):
        """Initialize a ArcTree.

ArcTree()
        """
        maptree.PointMapTree.__init__(self, Arc)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Arc object held in a ArcTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A Arc object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Arc object is found in the ArcTree within the
tolerance, the index of that object in the ArcTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if not isinstance(_o, Arc):
            raise TypeError, "Invalid type for searching in ArcTree: " + `obj`
        _arc = maptree.PointMapTree.binscan(self, _o, tol)
        return _arc

    def scan(self, obj, tol=tolerance.TOL):
        """Find a Arc object held in a ArcTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: An Arc object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Arc object is found in the ArcTree within the
tolerance, the index of that object in the ArcTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if not isinstance(_o, Arc):
            raise TypeError, "Invalid type for searching in ArcTree: " + `obj`
        _arc = maptree.PointMapTree.scan(self, _o, tol)
        return _arc

    def atCoords(self, x, y, rad, sa, ea):
        """Find an Arc at a specific location and some radius and angles.

atCoords(x, y, rad, sa, ea)

The arguments for this method are:

x: A float giving the center "x" coordinate.
y: A float giving the center "y" coordinate.
rad: A float giving radius of the Arc.
sa: A float giving the start angle of the Arc.
ea: A float giving the end angle of the Arc.

This method returns a Circle if there is one in the tree
at the specified location and with the desired radius.
Otherwise this method returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _rad = rad
        if not isinstance(_rad, float):
            _rad = float(rad)
        _sa = sa
        if not isinstance(_sa, float):
            _sa = float(sa)
        _ea = ea
        if not isinstance(_ea, float):
            _ea = float(ea)
        if _rad < 1e-10:
            raise ValueError, "Invalid radius: " + `_rad`
        if self.isModified():
            raise ValueError, "CircleTree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi: # use bisection to find a suitable arc ...
            _mid = (_hi + _lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _arc = self[_mid]
            _ax, _ay = _arc.getCenter().getCoords()
            _ar = _arc.getRadius()
            if abs(_ax - _x) < 1e-10:
                if abs(_ay - _y) < 1e-10:
                    if abs(_ar - _rad) < 1e-10:
                        _asa = _arc.getStartAngle()
                        if abs(_asa - _sa) < 1e-10:
                            _aa = _arc.getAngle()
                            _angle = _ea - _sa
                            if _angle < 0.0:
                                _angle = _angle + 360.0
                            if abs(_aa - _angle) < 1e-10:
                                _ro = _arc
                                break
                            elif _angle < _aa:
                                _lo = _mid + 1
                            else:
                                _hi = _mid
                        elif _sa < _asa:
                            _lo = _mid + 1
                        else:
                            _hi = _mid
                    elif _rad < _ar:
                        _lo = _mid + 1
                    else:
                        _hi = _mid
                elif _y < (_ay - _ar):
                    _lo = _mid + 1
                else:
                    _hi = _mid
            elif _x < (_ax - _ar):
                _lo = _mid + 1
            else:
                _hi = _mid
        return _ro

#
# Quadtree Arc storage
#

class ArcQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, Arc):
            raise TypeError, "Invalid Arc object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        # _axmin = _x - _r
        # _axmax = _x + _r
        # _aymin = _y - _r
        # _aymax = _y + _r
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _axmin - 1.0
            _ymin = _aymin - 1.0
            _xmax = _axmax + 1.0
            _ymax = _aymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _axmin < _xmin:
                _xmin = _axmin - 1.0
                _resize = True
            if _axmax > _xmax:
                _xmax = _axmax + 1.0
                _resize = True
            if _aymin < _ymin:
                _ymin = _aymin - 1.0
                _resize = True
            if _aymax > _ymax:
                _ymax = _aymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, ArcQuadtree.moveArc)

    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _xmin = _ymin = _xmax = _ymax = None
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        # _axmin = _x - _r
        # _axmax = _x + _r
        # _aymin = _y - _r
        # _aymax = _y + _r
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # arc may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 5:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _sa = args[3]
        if not isinstance(_sa, float):
            _sa = float(args[3])
        _ea = args[4]
        if not isinstance(_ea, float):
            _ea = float(args[4])
        _t = tolerance.TOL
        if _alen > 5:
            _t = tolerance.toltest(args[5])
        if not len(self):
            return None
        _axmin = _x - _r
        _axmax = _x + _r
        _aymin = _y - _r
        _aymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _adict = {}
        _bailout = False
        _arc = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < (_xmid - _t): # arc on left side
                    _ne = _se = False
                if _axmin > (_xmid + _t): # arc on right side
                    _nw = _sw = False
                if _aymax < (_ymid - _t): # arc below
                    _nw = _ne = False
                if _aymin > (_ymid + _t): # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _a in _node.getObjects():
                    _aid = id(_a)
                    if _aid not in _adict:
                        _ax, _ay = _a.getCenter().getCoords()
                        if ((abs(_ax - _x) < _t) and (abs(_ay - _y) < _t)):
                            _ar = _a.getRadius()
                            if abs(_ar - _r) < _t:
                                _asa = _a.getStartAngle()
                                _aea = _a.getEndAngle()
                                if abs(_asa - _sa) < 1e-10 and abs(_aea - _ea) < 1e-10:
                                    _bailout = True
                                    _arc = _a
                                break
                        _adict[_aid] = True
            if _bailout:
                break
        return _arc

    def moveArc(self, obj, *args):
        if obj not in self:
            raise ValueError, "Arc not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 5:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _sa = args[3]
        if not isinstance(_sa, float):
            _sa = float(args[3])
        _ea = args[4]
        if not isinstance(_ea, float):
            _ea = float(args[4])
        _axmin = _x - _r
        _axmax = _x + _r
        _aymin = _y - _r
        _aymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # arc may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _arc = _tsep = None
        _bailout = False
        _adict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _a in _node.getObjects():
                    _aid = id(_a)
                    if _aid not in _adict:
                        _ap = _a.mapCoords(_x, _y, _t)
                        if _ap is not None:
                            _ax, _ay = _ap
                            _sep = math.hypot((_ax - _x), (_ay - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _arc = _a
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _arc = _a
                            if _sep < 1e-10 and _arc is not None:
                                _bailout = True
                                break
            if _bailout:
                break
        return _arc

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _arcs = []
        if not len(self):
            return _arcs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _adict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _arc in _node.getObjects():
                    _aid = id(_arc)
                    if _aid not in _adict:
                        if _arc.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _arcs.append(_arc)
                        _adict[_aid] = True
        return _arcs
