#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# horizontal construction lines
#

import warnings

import conobject
import tolerance
import point
import util
import maptree
import quadtree

class HCLine(conobject.ConstructionObject):
    """A base class for horizontal construction lines.
    """
    messages = {
        'moved' : True,
        'keypoint_changed' : True
    }
    def __init__(self, p):
        """Instantiate an HCLine object.

HCLine(p)
        """
        _p = p
        if not isinstance(p, point.Point):
            _p = point.Point(p)
        conobject.ConstructionObject.__init__(self)
        self.__keypoint = _p
        _p.storeUser(self)
        _p.connect('moved', self, HCLine.movePoint)

    def __eq__(self, obj):
        """Compare one HCLine to another for equality.
        """
        if not isinstance(obj, HCLine):
            return False
        if obj is self:
            return True
        if abs(self.getLocation().y - obj.getLocation().y) < 1e-10:
            return True
        return False

    def __ne__(self, obj):
        """Compare one HCLine to another for inequality.
        """
        if not isinstance(obj, HCLine):
            return True
        if obj is self:
            return False
        if abs(self.getLocation().y - obj.getLocation().y) < 1e-10:
            return False
        return True

    def __str__(self):
        _x, _y = self.getLocation().getCoords()
        return "Horizontal Construction Line at y = %g" % self.__keypoint.y

    def getLocation(self):
        return self.__keypoint

    def setLocation(self, p):
        if self.isLocked():
            raise RuntimeError, "Setting keypoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid keypoint: " + `p`
        _kp = self.__keypoint
        if p is not _kp:
            _y = _kp.y
            _kp.disconnect(self)
            self.__keypoint = p
            self.sendMessage('keypoint_changed', _kp)
            p.connect('moved', self, HCLine.movePoint)
            p.storeUser(self)
            if abs(_y - p.y) > 1e-10:
                self.sendMessage('moved', p.x, _y)
            self.sendMessage('modified')

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the HCLine to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required argument:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the HCLine. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _hy = self.__keypoint.y
        if abs(_hy - _y) < _t:
            return _x, _hy
        return None

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the HCLine to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the HCLine. If the distance between the actual
Point and the Point used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        warnings.warn("mapPoint() is deprecated - use mapCoords()",
                      DeprecationWarning, stacklevel=2)
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _xp, _yp = _p.getCoords()
        return self.mapCoords(_xp, _yp, _t)

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Return whether or not an HCLine passes through a region.

inRegion(xmin, ymin, xmax, ymax)

The first four arguments define the boundary. The method
will return True if the HCLine falls between ymin and ymax.
Otherwise the function will return False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _y = self.__keypoint.y
        return _ymin < _y < _ymax

    def move(self, dx, dy):
        """Move a HCLine

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked() or self.__keypoint.isLocked():
            raise RuntimeError, "Moving HCLine not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__keypoint.getCoords()
            self.mute()
            self.__keypoint.move(_dx, _dy)
            self.unmute()
            self.sendMessage('moved', _x, _y)
            self.modified()

    def movePoint(self, p, *args):
        _plen = len(args)
        if _plen < 2:
            raise ValueError, "Invalid argument count: %d" % _plen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        if p is not self.__keypoint:
            raise ValueError, "Invalid point for HCLine::movePoint()" + `p`
        if abs(p.y - _y) > 1e-10:
            self.sendMessage('moved', _x, _y)
            self.modified()
            
    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _y = self.__keypoint.y
        if _ymin < _y < _ymax:
            return _xmin, _y, _xmax, _y
        return None

    def sendsMessage(self, m):
        if m in HCLine.messages:
            return True
        return conobject.ConstructionObject.sendsMessage(self, m)

class old_HCLine(conobject.Spcline):
    """A class for Horizontal construction lines.

A HCLine object is derived from an Spcline,so it shares
all the attributes and methods of that object.

A HCLine has the following addtional methods:

mapPoint(): Test if a point is within some distance to a HCLine.
mapCoords(): Test if a coordinate pair is within some distance to a HCLine.
inRegion(): Return whether or not a HCLine passes through a bounded region.
clone(): Return an identical copy of a HCLine.
    """

    # class methods

    def __init__(self, p):
        """Initialize an HCLine object.

HCLine(p)

The single argument should be a Point object.
        """
        conobject.Spcline.__init__(self, p)

    def __zeq__(self, obj):
        """Compare one HCLine to another for equality.
        """
        if not isinstance(obj, HCLine):
            raise TypeError, "Invalid comparison to %s" % type(obj)
        _val = False
        if abs(self.getLocation().y - obj.getLocation().y) < 1e-10:
            _val = True
        return _val

    def __zne__(self, obj):
        """Compare one HCLine to another for inequality.
        """
        if not isinstance(obj, HCLine):
            raise TypeError, "Invalid comparison to %s" % type(obj)
        _val = True
        if abs(self.getLocation().y - obj.getLocation().y) < 1e-10:
            _val = False
        return _val

    def __cmp__(self, obj):
        """Compare one HCLine object to another or to a Point.

The comparison of two HCLines is simple - the "larger"
of the two is the one with the greater "y" coordinate
value of their Point. If both values are within the
tolerance, the lines are equivalent.

Comparing an HCLine to a Point is equally simple.
If the "y" coordinate of the HCLine is greater than
the Point, return 1. If it is less, return -1. Otherwise,
return 0.
        """
        if not isinstance(obj, (point.Point, HCLine)):
            raise TypeError, "Invalid object for HCLine comparison: " + `obj`
        _xs, _ys = self.getLocation().getCoords()
        if isinstance(obj, HCLine):
            _x, _y = obj.getLocation().getCoords()
        else:
            _x, _y = obj.getCoords()
        if abs(_ys - _y) < 1e-10:
            _value = 0
        elif _ys > _y:
            _value = 1
        else:
            _value = -1
        return _value

    def __str__(self):
        _x, _y = self.getLocation().getCoords()
        return "Horizontal construction line at y = %g" % _y

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the HCLine to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required argument:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the HCLine. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _hx, _hy = self.getLocation().getCoords()
        _diff = abs(_y - _hy)
        _ap = None
        if _diff < _t or abs(_diff - _t) < 1e-10:
            _ap = point.Point(_x, _hy)
        return _ap

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the HCLine to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the HCLine. If the distance between the actual
Point and the Point used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _xp, _yp = _p.getCoords()
        return self.mapCoords(_xp, _yp, _t)

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Return whether or not an HCLine passes through a region.

inRegion(xmin, ymin, xmax, ymax)

The first four arguments define the boundary. The method
will return True if the HCLine falls between ymin and ymax.
Otherwise the function will return False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        return _ymin < _y < _ymax

    def clone(self):
        """Return a copy of a HCLine.

clone()
        """
        _loc = self.getLocation().clone()
        return HCLine(_loc)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        _coords = None
        if _ymin < _y < _ymax:
            _coords = (_xmin, _y, _xmax, _y)
        return _coords
    
class HCLineTree(maptree.PointMapTree):
    """A Tree class specific for HCLine objects.

This class is based on the PointMapTree class, and
holds only HCLine objects.
    """
    def __init__(self):
        """Initialize a HCLineTree.

HCLineTree()
        """
        maptree.PointMapTree.__init__(self, HCLine)

    def mapPoint(self, p, tol=tolerance.TOL):
        """See if an object in the HCLineTree exists at a Point.

mapPoint(p [, tol])

This method has one required parameter:

p: Either a Point object or a tuple of two-floats

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if the
     point is close enough to the objects held in the Tree.

The function returns a list of at tuples, each of the form

(obj, pt)

obj: Object that the Point is mapped to
pt: The projected Point on the object

The list will contain at most one tuple.
        """
        _p = p
        if not isinstance(_p, (point.Point, tuple)):
            raise TypeError, "Invalid type for searching in PointMapTree: " + `_p`
        if isinstance(_p, tuple):
            _x, _y = util.tuple_to_two_floats(_p)
            _p = point.Point(_x, _y)
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "HCLineTree in modified state."
        _px, _py = _p.getCoords()
        _objlist = []
        _scanlist = []
        _lo = 0
        _hi = len(self)
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _hcl = self[_mid]
            _hx, _hy = _hcl.getLocation().getCoords()
            if _py > _hy + _t + 1e-10:
                _lo = _mid + 1
            elif _py < _hy - _t - 1e-10:
                _hi = _mid
            else:
                _pt = _hcl.mapPoint(_p, _t)
                if _pt is not None:
                    _objlist.append((_hcl, _pt))
                break
        return _objlist

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a HCLine object held in a HCLineTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A HCLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     HCLine objects are close enough to be considered equal.

If a HCLine object is found in the HCLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, HCLine):
            raise TypeError, "Invalid type for searching in HCLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "HCLineTree in modified state."
        _lo = _mid = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _hcl = self[_mid]
            _res = cmp(obj, _hcl)
            if _res == 0:
                _ro = _hcl
                break
            elif _res == 1:
                _lo = _mid + 1
            else:
                _hi = _mid
        if _ro is None:
            _x, _y = obj.getLocation().getCoords()
            _ymin = _y - _t - 1e-10
            for _i in range(_mid, -1, -1):
                _hcl = self[_i]
                _hx, _hy = _hcl.getLocation().getCoords()
                if _hy < _ymin:
                    break
                _ydiff = abs(_hy - _y)
                if _ydiff < _t or abs(_ydiff - _t) < 1e-10:
                    _ro = _hcl
                    break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a HCLine object held in a HCLineTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: An HCLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     HCLine object are close enough to be considered equal.

If a HCLine object is found in the HCLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, HCLine):
            raise TypeError, "Invalid type for searching in HCLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "HCLineTree in modified state."
        _ro = None
        _x, _y = obj.getLocation().getCoords()
        _ymax = _y + _t + 1e-10
        for _sobj in self:
            _sx, _sy = _sobj.getLocation().getCoords()
            if _sy > _ymax:
                break
            _ydiff = abs(_sy - _y)
            if _ydiff < _t or abs(_ydiff - _t) < 1e-10:
                _ro = _sobj
                break
        return _ro

    def atCoords(self, y):
        """Find a HCLine in the HCLineTree at a certain location.

atCoords(y)

This method has one required argument:

y: A float value representing the "y" coordinate

If there is a HCLine in the HCLineTree at the designated
coordinates, it is returned. Otherwise this method returns None.

This method is similar to the binscan() method.
        """
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        if self.isModified():
            raise ValueError, "HCLineTree in modified state."
        _scanlist = []
        _lo = 0
        _hi = len(self)
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _hcl = self[_mid]
            _px, _py = _hcl.getLocation().getCoords()
            if abs(_py - _y) < 1e-10:
                _ro = _hcl
                break
            elif _py < _y:
                _lo = _mid + 1
            else:
                _hi = _mid
        return _ro

#
#
#

def intersect_region(hcl, xmin, ymin, xmax, ymax):
    if not isinstance(hcl, HCLine):
        raise TypeError, "Invalid HCLine: " + str(hcl)
    _xmin = xmin
    if not isinstance(_xmin, float):
        _xmin = float(xmin)
    _ymin = ymin
    if not isinstance(_ymin, float):
        _ymin = float(ymin)
    _xmax = xmax
    if not isinstance(_xmax, float):
        _xmax = float(xmax)
    if _xmax < _xmin:
        raise ValueError, "Illegal values: xmax < xmin"
    _ymax = ymax
    if not isinstance(_ymax, float):
        _ymax = float(ymax)
    if _ymax < _ymin:
        raise ValueError, "Illegal values: ymax < ymin"
    _x, _y = hcl.getLocation().getCoords()
    _x1 = _y1 = _x2 = _y2 = None
    if _ymin < _y < _ymax:
        _x1 = _xmin
        _y1 = _y
        _x2 = _xmax
        _y2 = _y
    return _x1, _y1, _x2, _y2

#
# Quadtree HCLine storage
#

class HCLineQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, HCLine):
            raise TypeError, "Invalid HCLine object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getLocation().getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _ymin = _bounds[1]
            _ymax = _bounds[3]
            if _node.hasSubnodes():
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _y < _ymid: # hcline below
                    _nw = _ne = False
                else: # hcline above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if (not (_y < _ymin) and (_y < _ymax)):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, HCLineQuadtree.moveHCLine)
        
    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getLocation().getCoords()
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _pdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _ymin = _bounds[1]
            _ymax = _bounds[3]
            if _node.hasSubnodes():
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _y < _ymid: # hcline below
                    _nw = _ne = False
                else: # hcline above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if (not (_y < _ymin) and (_y < _ymax)):
                    _node.delObject(obj)
                    _parent = _node.getParent()
                    if _parent is not None:                
                        _pid = id(_parent)
                        if _pid not in _pdict:
                            _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _y = args[0]
        if not isinstance(_y, float):
            _y = float(args[0])
        _t = tolerance.TOL
        if _alen > 1:
            _t = tolerance.toltest(args[1])
        if not len(self):
            return None
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _hcl = _tsep = None
        _hdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _ymin = _bounds[1]
            _ymax = _bounds[3]
            if (_ymin - _t) < _y < (_ymax + _t):
                if _node.hasSubnodes():
                    for _subnode in _node.getSubnodes():
                        _bounds = _node.getBoundary()
                        _ymin = _bounds[1]
                        _ymax = _bounds[3]
                        if (_ymin - _t) < _y < (_ymax + _t):
                            _nodes.append(_subnode)
                else:
                    for _h in _node.getObjects():
                        _hid = id(_h)
                        if _hid not in _hdict:
                            _yh = _h.getLocation().y
                            _ysep = abs(_yh - _y)
                            if _ysep < _t:
                                if _hcl is None:
                                    _hcl = _h
                                    _tsep = _ysep
                                else:
                                    if _ysep < _tsep:
                                        _hcl = _h
                                        _tsep = _ysep
                            _hdict[_hid] = True
        return _hcl

    def moveHCLine(self, obj, *args):
        if obj not in self:
            raise ValueError, "HCLine not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _ymin = _bounds[1]
            _ymax = _bounds[3]
            if _node.hasSubnodes():
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _y < _ymid: # hcline below
                    _nw = _ne = False
                else: # hcline above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # hcline may not be in node
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)
    
    def getClosest(self, x, y, tol=tolerance.TOL):
        return self.find(y, tol)

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _hcls = []        
        if not len(self):
            return _hcls
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _hdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _bounds = _subnode.getBoundary()
                    _bmin = _bounds[1]
                    _bmax = _bounds[3]
                    if ((_bmin > _ymax) or (_bmax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _hcl in _node.getObjects():
                    _hid = id(_hcl)
                    if _hid not in _hdict:
                        if _hcl.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _hcls.append(_hcl)
                        _hdict[_hid] = True
        return _hcls
