#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# The Layer class
#

import warnings
import sys

import point
import segment
import circle
import arc
import hcline
import vcline
import acline
import cline
import ccircle
import segjoint
import leader
import polyline
import text
import dimension
import dimtrees
import tolerance
import baseobject
import entity

class Layer(entity.Entity):
    """The Layer class.

A Layer object holds all the various entities that can be
in a drawing. Each layer can have sublayers, and there is
no limit to the depth of the sublayering.

A Layer object has several attributes:

name: The Layer's name
parent: The parent Layer of the Layer
scale: The scale factor for object contained in the Layer
visibility: A flag to show the Layer's contents or not

A Layer object has the following methods:

{get/set}Name(): Get/Set the Layer's name.
{get/set}Parent(): Get/Set the Layer's parent.
hasChildren(): Test if the Layer has children.
getChildren(): Return the children of the Layer.
getEntities(): Return all of a type of entity in the Layer.
{add/del}Child(): Add/Remove a new child Layer to the Layer.
{add/del}Object(): Store/Remove a Point, Segment, etc. in the Layer.
findObject(): Return an object in the layer equivalent to a test object.
hasPoint(): See if the Layer has a Point object at a location.
hasPointAt(): Find a Point in the Layer.
hasCircleAt(): Find a Circle in the Layer.
mapPoint(): See if a non-Point object in the layer crosses some location.
addLayer(): Add the contents of one layer to another.
{get/set}Visibility(): Get/Set the visibility of the Layer.
getBoundary(): Find the maximum and minimum coordinates of the Layer.
objsInRegion(): Return all the objects in the Layer that can be seen
                within some view.
update(): Check that all objects in the layer are stored properly.
    """

    messages = {
        'attribute_changed' : True,
        'added_object' : True,
        'deleted_object' : True,
        }
    def __init__(self, name=None):
        """Initializee a Layer.

Layer(name)

The 'name' argument is optional. It should be a string if used,
otherwise a default name "Layer" is assigned to the layer.
        """
        _n = name
        if _n is None:
            _n = u'Layer'
        if not isinstance(_n, unicode):
            if isinstance(name, str):
                _n = unicode(name)
            else:
                raise TypeError, "Invalid layer name: "+ str(name)
        entity.Entity.__init__(self)
        self.__name = _n
        self.__parent = None
        self.__points = point.PointQuadtree()
        self.__segments = segment.SegmentQuadtree()
        self.__circles = circle.CircleQuadtree()
        self.__arcs = arc.ArcQuadtree()
        self.__hclines = hcline.HCLineQuadtree()
        self.__vclines = vcline.VCLineQuadtree()
        self.__aclines = acline.ACLineQuadtree()
        self.__clines = cline.CLineQuadtree()
        self.__ccircles = ccircle.CCircleQuadtree()
        self.__chamfers = []
        self.__fillets = []
        self.__leaders = leader.LeaderQuadtree()
        self.__polylines = polyline.PolylineQuadtree()
        self.__textblocks = []
        self.__ldims = dimtrees.LDimQuadtree()
        self.__hdims = dimtrees.HDimQuadtree()
        self.__vdims = dimtrees.VDimQuadtree()
        self.__rdims = dimtrees.RDimQuadtree()
        self.__adims = dimtrees.ADimQuadtree()
        self.__children = []
        self.__visibility = True
        self.__scale = 1.0
        #
        # self.__objects keeps a reference to all objects stored in the layer
        #
        self.__objects = {}

    def __str__(self):
        if self.__parent is None:
            string = "Layer: %s [No Parent]" % self.__name
        else:
            string = "Layer: %s; Parent: %s" % (self.__name, self.__parent.__name)
        return string

    def __contains__(self, obj):
        """Find an object in the Layer.

This method permits the use of 'in' for test conditions.

if obj in layer:
    ....

This function tests for Point, Segment, Circle, etc. It returns
True if there is an equivalent object held in the Layer. Otherwise
the function returns False.
        """
        _seen = False
        if id(obj) in self.__objects:
            _seen = True
        if not _seen:
            self.update()
            if isinstance(obj, point.Point):
                _x, _y = obj.getCoords()
                _seen = (self.__points.find(_x, _y) is not None)
            elif isinstance(obj, segment.Segment):
                _p1, _p2 = obj.getEndpoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _seen = (self.__segments.find(_x1, _y1, _x2, _y2) is not None)
            elif isinstance(obj, arc.Arc):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _sa = obj.getStartAngle()
                _ea = obj.getEndAngle()
                _seen = (self.__arcs.find(_x, _y, _r, _sa, _ea) is not None)
            elif isinstance(obj, circle.Circle):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _seen = (self.__circles.find(_x, _y, _r) is not None)
            elif isinstance(obj, hcline.HCLine):
                _y = obj.getLocation().y
                _seen = (self.__hclines.find(_y) is not None)
            elif isinstance(obj, vcline.VCLine):
                _x = obj.getLocation().x
                _seen = (self.__vclines.find(_x) is not None)
            elif isinstance(obj, acline.ACLine):
                _x, _y = obj.getLocation().getCoords()
                _angle = obj.getAngle()
                _seen = (self.__aclines.find(_x, _y, _angle) is not None)
            elif isinstance(obj, cline.CLine):
                _p1, _p2 = obj.getKeypoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _seen = (self.__clines.find(_x1, _y1, _x2, _y2) is not None)
            elif isinstance(obj, ccircle.CCircle):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _seen = (self.__ccircles.find(_x, _y, _r) is not None)
            elif isinstance(obj, segjoint.Fillet):
                _seen = obj in self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                _seen = obj in self.__chamfers
            elif isinstance(obj, leader.Leader):
                _p1, _p2, _p3 = obj.getPoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _x3, _y3 = _p3.getCoords()
                _seen = (self.__leaders.find(_x1, _y1, _x2, _y2,
                                             _x3, _y3) is not None)
            elif isinstance(obj, polyline.Polyline):
                _coords = []
                for _pt in obj.getPoints():
                    _coords.extend(_pt.getCoords())
                _seen = (self.__polylines.find(_coords) is not None)
            elif isinstance(obj, text.TextBlock):
                _seen = obj in self.__textblocks
            elif isinstance(obj, dimension.HorizontalDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _seen = (self.__hdims.find(_l1, _p1, _l2, _p2) is not None)
            elif isinstance(obj, dimension.VerticalDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _seen = (self.__vdims.find(_l1, _p1, _l2, _p2) is not None)
            elif isinstance(obj, dimension.LinearDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _seen = (self.__ldims.find(_l1, _p1, _l2, _p2) is not None)
            elif isinstance(obj, dimension.RadialDimension):
                _l1 = obj.getDimLayer()
                _c1 = obj.getDimCircles()
                _seen = (self.__rdims.find(_l1, _c1) is not None)
            elif isinstance(obj, dimension.AngularDimension):
                _vl, _l1, _l2 = obj.getDimLayers()
                _vp, _p1, _p2 = obj.getDimPoints()
                _dim = self.__adims.find(_vl, _vp, _l1, _p1, _l2, _p2)
                _seen = dim is not None
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
        return _seen

    def clear(self):
        """Remove all the entities stored in this layer

clear()
        """
        print "clearing layer: " + self.__name
        self.__points.delObjects()
        self.__segments.delObjects()
        self.__circles.delObjects()
        self.__arcs.delObjects()
        self.__hclines.delObjects()
        self.__vclines.delObjects()
        self.__aclines.delObjects()
        self.__clines.delObjects()
        self.__ccircles.delObjects()
        del self.__chamfers[:]
        del self.__fillets[:]
        self.__leaders.delObjects()
        self.__polylines.delObjects()
        del self.__textblocks[:]
        self.__ldims.delObjects()
        self.__hdims.delObjects()
        self.__vdims.delObjects()
        self.__rdims.delObjects()
        self.__adims.delObjects()
        self.__visibility = True
        self.__scale = 1.0
        self.__objects.clear()

    def free(self):
        """Disconnect a layer from its parent.

free()
        """
        print "freeing layer: " + self.__name
        _parent = self.__parent
        if _parent is not None:
            _parent.delChild(self)

    def getName(self):
        """Return the name of the Layer.

getName()
        """
        return self.__name

    def setName(self, name):
        """Set the name of the Layer.

setName(name)

The new must be a string, and cannot be None.
        """
        _n = name
        if _n is None:
            raise ValueError, "Layers must have a name."
        if not isinstance(_n, unicode):
            if isinstance(_n, str):
                _n = unicode(_n)
            else:
                raise TypeError, "Invalid name: " + `_n`
        _on = self.__name
        if _on != _n:
            self.__name = _n
            self.sendMessage('attribute_changed', 'name', _on)

    name = property(getName, setName, None, "Layer name.")

    def addLayer(self, l):
        """Add one Layer to another.

addLayer(l)

Add the contents of the Layer object "l" into the Layer. Only
the objects not found in the current Layer are added.
        """
        warnings.warn("addLayer() is deprecated - will be replaced later ...",
                      DeprecationWarning, stacklevel=2)
        if not isinstance(l, Layer):
            raise TypeError, "Invalid Layer addition object: " + `l`
        _objmap = {}
        self.update()
        l.update()
        for _obj in l.getLayerEntities("point"):
            _sobj = self.__point_tree.binscan(_obj)
            if _sobj is None:
                self.addObject(_obj)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("segment"):
            _sobj = self.__segment_tree.binscan(_obj)
            if _sobj is None:
                _p1, _p2 = _obj.getEndpoints()
                if _p1 in _objmap:
                    _p1 = _objmap[_p1]
                if _p2 in _objmap:
                    _p2 = _objmap[_p2]
                _ns = _obj.clone()
                _ns.setP1(_p1)
                _ns.setP2(_p2)
                _ns.reset()
                self.addObject(_ns)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("circle"):
            _sobj = self.__circle_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                _nc = _obj.clone()
                _nc.setCenter(_cp)
                _nc.reset()
                self.addObject(_nc)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("arc"):
            _sobj = self.__arc_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                _na = _obj.clone()
                _na.setCenter(_cp)
                _na.reset()
                self.addObject(_na)
            else:
                _objmap[_obj] = _sobj
        for _obj in l.getLayerEntities("hcline"):
            _sobj = self.__hcline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(hcline.HCLine(_lp))
        for _obj in l.getLayerEntities("vcline"):
            _sobj = self.__vcline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(vcline.VCLine(_lp))
        for _obj in l.getLayerEntities("acline"):
            _sobj = self.__acline_tree.binscan(_obj)
            if _sobj is None:
                _lp = _obj.getLocation()
                if _lp in _objmap:
                    _lp = _objmap[_lp]
                self.addObject(acline.ACLine(_lp, _obj.getAngle()))
        for _obj in l.getLayerEntities("cline"):
            _sobj = self.__cline_tree.binscan(_obj)
            if _sobj is None:
                _p1, _p2 = _obj.getKeypoints()
                if _p1 in _objmap:
                    _p1 = _objmap[_p1]
                if _p2 in _objmap:
                    _p2 = _objmap[_p2]
                self.addObject(cline.CLine(_p1, _p2))
        for _obj in l.getLayerEntities("ccircle"):
            _sobj = self.__ccircle_tree.binscan(_obj)
            if _sobj is None:
                _cp = _obj.getCenter()
                if _cp in _objmap:
                    _cp = _objmap[_cp]
                self.addObject(ccircle.CCircle(_cp, _obj.getRadius()))
        for _obj in l.getLayerEntities("chamfer"):
            _s1, _s2 = _obj.getSegments()
            if _s1 in _objmap:
                _s1 = _objmap[_s1]
            if _s2 in _objmap:
                _s2 = _objmap[_s2]
            _nc = segjoint.Chamfer(_s1, _s2, _obj.getLength(),
                                   _obj.getStyle(), _obj.getLinetype(),
                                   _obj.getColor(), _obj.getThickness())
            if _nc not in self.__chamfers:
                self.addObject(_nc)
        for _obj in l.getLayerEntities("fillet"):
            _s1, _s2 = _obj.getSegments()
            if _s1 in _objmap:
                _s1 = _objmap[_s1]
            if _s2 in _objmap:
                _s2 = _objmap[_s2]
            _nf = segjoint.Fillet(_s1, _s2, _obj.getRadius(),
                                  _obj.getStyle(), _obj.getLinetype(),
                                  _obj.getColor(), _obj.getThickness())
            if _nf not in self.__fillets:
                self.addObject(_nf)
        for _obj in l.getLayerEntities("leader"):
            if _obj not in self.__leaders:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("polyline"):
            if _obj not in self.__polylines:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("text"):
            if _obj not in self.__textblocks:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("linear_dimension"):
            if _obj not in self.__linear_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("horizontal_dimension"):
            if _obj not in self.__horiz_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("vertical_dimension"):
            if _obj not in self.__vert_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("radial_dimension"):
            if _obj not in self.__radial_dimensions:
                self.addObject(_obj)
        for _obj in l.getLayerEntities("angular_dimension"):
            if _obj not in self.__angular_dimensions:
                self.addObject(_obj)

    def addObject(self, obj):
        """Add an object to this Layer.

addObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, Polyline, or Dimension. Anything else raises
a TypeError exception.
        """
        if self.isLocked():
            raise RuntimeError, "Adding entity not allowed - layer locked."
        if id(obj) in self.__objects:
            return
        self.update()
        if isinstance(obj, point.Point):
            self._addPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._addSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._addArc(obj)
        elif isinstance(obj, circle.Circle):
            self._addCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._addHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._addVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._addACLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._addCCircle(obj)
        elif isinstance(obj, cline.CLine):
            self._addCLine(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._addChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._addFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._addLeader(obj)
        elif isinstance(obj, polyline.Polyline):
            self._addPolyline(obj)
        elif isinstance(obj, text.TextBlock):
            self._addTextBlock(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._addAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._addRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._addHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._addVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._addLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for storage: " + `obj`
        self.sendMessage('added_object', obj)

    def _addPoint(self, p):
        """Add a Point object to the Layer.

_addPoint(p)

This method is private to the Layer object.
        """
        _x, _y = p.getCoords()
        if self.__points.find(_x, _y) is None:
            self.__points.addObject(p)
            self.__objects[id(p)] = p

    def _addSegment(self, s):
        """Add a Segment object to the Layer.

_addSegment(s)

The Segment object "s" is added to the Layer if there is not
already a Segment in the layer between the same locations.
The Segment endpoints p1 and p2 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _p1, _p2 = s.getEndpoints()
        if id(_p1) not in self.__objects:
            raise ValueError, "Segment p1 Point not found in Layer."
        if id(_p2) not in self.__objects:
            raise ValueError, "Segment p2 Point not found in Layer."
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        if self.__segments.find(_x1, _y1, _x2, _y2) is None:
            self.__segments.addObject(s)
            self.__objects[id(s)] = s
        else:
            _p1.freeUser(s)
            _p2.freeUser(s)

    def _addCircle(self, c):
        """Add a Circle object to the Layer.

_addCircle(c)

The Circle object "c" is added to the Layer if there is not
already a Circle at that location with the same radius. The
Circle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the layer object.
        """
        _cp = c.getCenter()
        if id(_cp) not in self.__objects:
            raise ValueError, "Circle center Point not found in Layer."
        _x, _y = _cp.getCoords()
        _r = c.getRadius()
        if self.__circles.find(_x, _y, _r) is None:
            self.__circles.addObject(c)
            self.__objects[id(c)] = c
        else:
            _cp.freeUser(c)

    def _addArc(self, a):
        """Add an Arc object to the Layer.

_addArc(a)

The Arc object "a" is added to the Layer if there is not
already a Arc at that location with the same radius, start
angle, and eng_angle. The Arc center must already be stored
in the Layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _cp = a.getCenter()
        if id(_cp) not in self.__objects:
            raise ValueError, "Arc center Point not found in Layer."
        _x, _y = _cp.getCoords()
        _r = a.getRadius()
        _sa = a.getStartAngle()
        _ea = a.getEndAngle()
        if self.__arcs.find(_x, _y, _r, _sa, _ea) is None:
            self.__arcs.addObject(a)
            self.__objects[id(a)] = a
            for _ex, _ey in a.getEndpoints():
                _lp = self.__points.find(_ex, _ey)
                if _lp is None:
                    _lp = point.Point(_ex, _ey)
                    self.addObject(_lp)
                _lp.storeUser(a)

    def _addHCLine(self, hcl):
        """Add an HCLine object to the Layer.

_addHCLine(hcl)

The HCLine object "hcl" is added to the Layer if there is not
already a HCLine at that location. The HCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private the the Layer object.
        """
        _lp = hcl.getLocation()
        if id(_lp) not in self.__objects:
            raise ValueError, "HCLine location Point not found in Layer."
        _y = _lp.y
        if self.__hclines.find(_y) is None:
            self.__hclines.addObject(hcl)
            self.__objects[id(hcl)] = hcl
        else:
            _lp.freeUser(hcl)

    def _addVCLine(self, vcl):
        """Add an VCLine object to the Layer.

_addVCLine(vcl)

The VCLine object "vcl" is added to the Layer if there is not
already a VCLine at that location. The VCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _lp = vcl.getLocation()
        if id(_lp) not in self.__objects:
            raise ValueError, "VCLine location Point not found in Layer."
        _x = _lp.x
        if self.__vclines.find(_x) is None:
            self.__vclines.addObject(vcl)
            self.__objects[id(vcl)] = vcl
        else:
            _lp.freeUser(vcl)

    def _addACLine(self, acl):
        """Add an ACLine object to the Layer.

_addACLine(acl)

The ACLine object "acl" is added to the Layer if there is not
already a ACLine at that location. The ACLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _lp = acl.getLocation()
        if id(_lp) not in self.__objects:
            raise ValueError, "ACLine location Point not found in Layer."
        _x, _y = _lp.getCoords()
        _angle = acl.getAngle()
        if self.__aclines.find(_x, _y, _angle) is None:
            self.__aclines.addObject(acl)
            self.__objects[id(acl)] = acl

    def _addCCircle(self, cc):
        """Add an CCircle object to the Layer.

_addCCircle(cc)

The CCircle object "cc" is added to the Layer if there is not
already a CCircle at that location and with the same radius.
The CCircle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the Layer object.
        """
        _cp = cc.getCenter()
        if id(_cp) not in self.__objects:
            raise ValueError, "CCircle center Point not found in Layer."
        _x, _y = _cp.getCoords()
        _r = cc.getRadius()
        if self.__circles.find(_x, _y, _r) is None:
            self.__ccircles.addObject(cc)
            self.__objects[id(cc)] = cc
        else:
            _cp.freeUser(cc)

    def _addCLine(self, cl):
        """Add an CLine object to the Layer.

_addCLine(cl)

The CLine object "cl" is added to the Layer if there is not
already a CLine at that location. The CLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        _p1, _p2 = cl.getKeypoints()
        if id(_p1) not in self.__objects:
            raise ValueError, "CLine p1 Point not found in Layer."
        if id(_p2) not in self.__objects:
            raise ValueError, "CLine p2 Point not found in Layer."
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        if self.__clines.find(_x1, _y1, _x2, _y2) is None:
            self.__clines.addObject(cl)
            self.__objects[id(cl)] = cl
        else:
            _p1.freeUser(cl)
            _p2.freeUser(cl)

    def _addChamfer(self, ch):
        """Add a Chamfer object to the Layer.

_addChamfer(ch)

The Chamfer object "ch" is added to the Layer if there is not
already a Chamfer found joining the two Segments. Both Segment
objects in the chamfer must also be stored in this Layer.

This method is private to the Layer object.
        """
        _s1, _s2 = ch.getSegments()
        if ch in self.__chamfers:
            _s1.freeUser(ch)
            _s2.freeUser(ch)
        else:
            if id(_s1) not in self.__objects:
                raise ValueError, "Chamfer s1 Segment not found in Layer."
            if id(_s2) not in self.__objects:
                raise ValueError, "Chamfer s2 Segment not found in Layer."
            _new_chamfer = True
            for _ch in self.__chamfers:
                if _ch == ch:
                    _new_chamfer = False
                    break
            if _new_chamfer:
                self.__objects[id(ch)] = ch
                self.__chamfers.append(ch)
            else:
                _s1.freeUser(ch)
                _s2.freeUser(ch)

    def _addFillet(self, f):
        """Add a Fillet object to the Layer.

_addFillet(f)

The Fillet object "f" is added to the Layer if there is not
already a Fillet found joining the two Segments. Both Segment
objects in the fillet must also be stored in this Layer.

This method is private to the Layer object.
        """
        _s1, _s2 = f.getSegments()
        if f in self.__fillets:
            _s1.freeUser(f)
            _s2.freeUser(f)
        else:
            if id(_s1) not in self.__objects:
                raise ValueError, "Fillet s1 Segment not found in Layer."
            if id(_s2) not in self.__objects:
                raise ValueError, "Fillet s2 Segment not found in Layer."
            _new_fillet = True
            for _f in self.__fillets:
                if _f == f:
                    _new_fillet = False
                    break
            if _new_fillet:
                self.__objects[id(f)] = f
                self.__fillets.append(f)
            else:
                _s1.freeUser(f)
                _s2.freeUser(f)

    def _addLeader(self, l):
        """Add a Leader object to the Layer.

_addLeader(l)

The Leader object "l" is added to the Layer if there is not
already a Leader in the layer between the same locations.
The Leader endpoints p1, p2, and p3 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        _p1, _p2, _p3 = l.getPoints()
        if id(_p1) not in self.__objects:
            raise ValueError, "Leader p1 Point not found in Layer."
        if id(_p2) not in self.__objects:
            raise ValueError, "Leader p2 Point not found in Layer."
        if id(_p3) not in self.__objects:
            raise ValueError, "Leader p3 Point not found in Layer."
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _x3, _y3 = _p3.getCoords()
        if self.__leaders.find(_x1, _y1, _x2, _y2, _x3, _y3) is None:
            self.__leaders.addObject(l)
            self.__objects[id(l)] = l
        else:
            _p1.freeUser(l)
            _p2.freeUser(l)
            _p3.freeUser(l)

    def _addPolyline(self, pl):
        """Add a Polyline object to the Layer.

_addPolyline(pl)

The Polyline object "pl" is added to the Layer if there is not
already a Polyline in the layer at the same locations. The Polyline
points _must_ be already stored in this layer, otherwise a ValueError
is raised.

This method is private to the Layer object.
        """
        _coords = []
        _pts = pl.getPoints()
        for _pt in _pts:
            if id(_pt) not in self.__objects:
                raise ValueError, "Polyline point not in layer: " + str(_pt)
            _coords.append(_pt.getCoords())
        if self.__polylines.find(_coords) is None:
            self.__polylines.addObject(pl)
            self.__objects[id(pl)] = pl
        else:
            for _pt in _pts:
                _pt.freeUser(pl)

    def _addTextBlock(self, tb):
        """Add a TextBlock object to the Layer.

_addTextBlock(tb)

The TextBlock object "tb" is added to the layer if there is not
already a TextBlock in the layer at the same location and with the
same text.
        """
        if tb not in self.__textblocks:
            self.__textblocks.append(tb)
            self.__objects[id(tb)] = tb

    def _addAngularDimension(self, adim):
        """Add an AngularDimension object to the Layer.

_addAngularDimension(adim)

This method is private to the Layer object.
        """
        _p1, _p2, _p3 = adim.getDimPoints()
        _l1, _l2, _l3 = adim.getDimLayers()
        _lp = _l1.findObject(_p1)
        if _lp is not _p1:
            raise ValueError, "Dimension Point p1 not found in layer"
        _lp = _l2.findObject(_p2)
        if _lp is not _p2:
            raise ValueError, "Dimension Point p2 not found in layer!"
        _lp = _l3.findObject(_p3)
        if _lp is not _p3:
            raise ValueError, "Dimension Point p3 not found in layer!"
        if self.__adims.find(_l1, _p1, _l2, _p2, _l3, _p3) is None:
            self.__objects[id(adim)] = adim
            self.__adims.addObject(adim)
        else:
            _p1.freeUser(adim)
            _p2.freeUser(adim)
            _p3.freeUser(adim)

    def _addRadialDimension(self, rdim):
        """Add a RadialDimension object to the Layer.

_addRadialDimension(rdim)

This method is private to the Layer object.
        """
        _dc = rdim.getDimCircle()
        _dl = rdim.getDimLayer()
        _lc = _dl.findObject(_dc)
        if _lc is not _dc:
            raise ValueError, "RadialDimension circular object not found in layer"
        if self.__rdims.find(_dl, _dc) is None:
            self.__objects[id(rdim)] = rdim
            self.__rdims.addObject(rdim)
        else:
            _dc.freeUser(rdim)

    def _addHorizontalDimension(self, hdim):
        """Add a HorizontalDimension object to the Layer.

_addHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        _l1, _l2 = hdim.getDimLayers()
        _p1, _p2 = hdim.getDimPoints()
        _lp = _l1.findObject(_p1)
        if _lp is not _p1:
            raise ValueError, "HorizontalDimension Point p1 not found in layer"
        _lp = _l2.findObject(_p2)
        if _lp is not _p2:
            raise ValueError, "HorizontalDimension Point p2 not found in layer"
        if self.__hdims.find(_l1, _p1, _l1, _p2) is None:
            self.__objects[id(hdim)] = hdim
            self.__hdims.addObject(hdim)
        else:
            _p1.freeUser(hdim)
            _p2.freeUser(hdim)

    def _addVerticalDimension(self, vdim):
        """Add a VerticalDimension object to the Layer.

_addVerticalDimension(vdim)

This method is private to the Layer object.
        """
        _l1, _l2 = vdim.getDimLayers()
        _p1, _p2 = vdim.getDimPoints()
        _lp = _l1.findObject(_p1)
        if _lp is not _p1:
            raise ValueError, "VerticalDimension Point p1 not found in layer"
        _lp = _l2.findObject(_p2)
        if _lp is not _p2:
            raise ValueError, "VerticalDimension Point p2 not found in layer"
        if self.__vdims.find(_l1, _p1, _l1, _p2) is None:
            self.__objects[id(vdim)] = vdim
            self.__vdims.addObject(vdim)
        else:
            _p1.freeUser(vdim)
            _p2.freeUser(vdim)

    def _addLinearDimension(self, ldim):
        """Add a LinearDimension object to the Layer.

_addLinearDimension(ldim)

This method is private to the Layer object.
        """
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        _lp = _l1.findObject(_p1)
        if _lp is not _p1:
            raise ValueError, "LinearDimension Point p1 not found in layer"
        _lp = _l2.findObject(_p2)
        if _lp is not _p2:
            raise ValueError, "LinearDimension Point p2 not found in layer"
        if self.__ldims.find(_l1, _p1, _l1, _p2) is None:
            self.__objects[id(ldim)] = ldim
            self.__ldims.addObject(ldim)
        else:
            _p1.freeUser(ldim)
            _p2.freeUser(ldim)

    def delObject(self, obj):
        """Remove an object from this Layer.

delObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, or Dimension. Anything else raises
a TypeError exception.
        """
        if self.isLocked():
            raise RuntimeError, "Deleting entity not allowed - layer locked."
        if id(obj) not in self.__objects:
            raise ValueError, "Object not found in layer: " + `obj`
        self.update()
        if isinstance(obj, point.Point):
            self._delPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._delSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._delArc(obj)
        elif isinstance(obj, circle.Circle):
            self._delCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._delHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._delVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._delACLine(obj)
        elif isinstance(obj, cline.CLine):
            self._delCLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._delCCircle(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._delChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._delFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._delLeader(obj)
        elif isinstance(obj, polyline.Polyline):
            self._delPolyline(obj)
        elif isinstance(obj, text.TextBlock):
            self._delTextBlock(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._delAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._delRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._delHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._delVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._delLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for removal: " + `obj`
        self.sendMessage('deleted_object', obj)

    def _delPoint(self, p):
        """Delete a Point from the Layer.

_delPoint(p)

This method is private to the Layer object.
        """
        _delete = True
        _users = p.getUsers()
        for _uref in _users:
            _user = _uref()
            if _user is not None:
                if not isinstance(_user, dimension.Dimension):
                    _delete = False
                    break
        if _delete:
            _extdims = {}
            for _uref in _users:
                _user = _uref()
                if _user is not None:
                    _uid = id(_user)
                    if isinstance(_user, dimension.Dimension):
                        if _uid not in self.__objects:
                            _extdims[_uid] = _user
                    else:
                        if _uid not in self.__objects:
                            print "point bound to non-layer object: " + `_user`
                            _delete = False
                            break
            if _delete:
                for _uref in _users:
                    _user = _uref()
                    if _user is not None:
                        if id(_user) not in _extdims:
                            self.delObject(_user)
                if len(_extdims):
                    _parent = self.getParent()
                    _top = self
                    while _parent is not None:
                        _top = _parent
                        _parent = _top.getParent()
                    _layers = []
                    _layers.append(_top)
                    while len(_layers):
                        _layer = _layers.pop(0)
                        _dimids = _extdims.keys()
                        for _dimid in _dimids:
                            _dim = _extdims[_dimid]
                            if _dim in _layer:
                                del _extdims[_dimid]
                                _layer.delObject(_dim)
                        if len(_extdims):
                            if _layer.hasChildren():
                                _layers.extend(_layer.getChildren())
                _pid = id(p)
                del self.__objects[_pid]
                self.__points.delObject(p)

    def _delSegment(self, s, all=True):
        """Delete a Segment from the Layer.

_delSegment(s[, all=True])

This method is private to the Layer object.
        """
        for _uref in s.getUsers(): # chamfers, fillets, or hatching
            _user = _uref()
            if _user is not None:
                self.delObject(_user) # restore other segments?
        _p1, _p2 = s.getEndpoints()
        assert id(_p1) in self.__objects, "Segment p1 Point not in objects"
        _p1.freeUser(s)
        assert id(_p2) in self.__objects, "Segment p2 Point not in objects"
        _p2.freeUser(s)
        del self.__objects[id(s)]
        self.__segments.delObject(s)
        if all:
            self.delObject(_p1) # remove possibly unused point _p1
            self.delObject(_p2) # remove possibly unused point _p2

    def _delCircle(self, c, all=True):
        """Delete a Circle from the Layer.

_delCircle(c[, all=True])

This method is private to the Layer object.
        """
        assert not isinstance(c, arc.Arc), "Arc in _delCircle()"
        _extdims = {}
        for _uref in c.getUsers(): # dimensions or hatching
            _user = _uref()
            if _user is not None:
                _uid = id(_user)
                if isinstance(_user, dimension.RadialDimension):
                    if _uid not in self.__objects:
                        _extdims[_uid] = _user
                    else:
                        self.delObject(_user)
        if len(_extdims):
            _parent = self.getParent()
            _top = self
            while _parent is not None:
                _top = _parent
                _parent = _top.getParent()
            _layers = []
            _layers.append(_top)
            while len(_layers):
                _layer = _layers.pop(0)
                _dimids = _extdims.keys()
                for _dimid in _dimids:
                    _dim = _extdims[_dimid]
                    if _dim in _layer:
                        del _extdims[_dimid]
                        _layer.delObject(_dim)
                if len(_extdims):
                    _layers.extend(_layer.getChildren())
        _cp = c.getCenter()
        assert id(_cp) in self.__objects, "Circle center point not in objects"
        _cp.freeUser(c)
        del self.__objects[id(c)]
        self.__circles.delObject(c)
        if all:
            self.delObject(_cp) # remove possibly unused point _cp

    def _delArc(self, a, all=True):
        """Delete an Arc from the Layer.

_delArc(a[, all=True])

This method is private to the Layer object.
        """
        _extdims = {}
        for _uref in a.getUsers(): # dimensions or hatching
            _user = _uref()
            if _user is not None:
                _uid = id(_user)
                if isinstance(_user, dimension.RadialDimension):
                    if _uid not in self.__objects:
                        _extdims[_uid] = _user
                    else:
                        self.delObject(_user)
        if len(_extdims):
            _parent = self.getParent()
            _top = self
            while _parent is not None:
                _top = _parent
                _parent = _top.getParent()
            _layers = []
            _layers.append(_top)
            while len(_layers):
                _layer = _layers.pop(0)
                _dimids = _extdims.keys()
                for _dimid in _dimids:
                    _dim = _extdims[_dimid]
                    if _dim in _layer:
                        del _extdims[_dimid]
                        _layer.delObject(_dim)
                if len(_extdims):
                    _layers.extend(_layer.getChildren())
        _cp = a.getCenter()
        assert id(_cp) in self.__objects, "Arc center point not in objects"
        _cp.freeUser(a)
        for _ep in a.getEndpoints():
            _p = self.find('point', _ep[0], _ep[1])
            assert _p is not None, "Arc endpoint not found in layer"
            assert id(_p) in self.__objects, "Arc endpoint not in objects"
            _p.freeUser(a)
            if all:
                self.delObject(_p) # remove possibly unused point _p
        del self.__objects[id(a)]
        self.__arcs.delObject(a)
        if all:
            self.delObject(_cp) # remove possibly unused point _cp

    def _delHCLine(self, hcl, all=True):
        """Remove a HCLine object from the Layer.

_delHCLine(hcl[, all=True])

This method is private to the Layer object.
        """
        _lp = hcl.getLocation()
        assert id(_lp) in self.__objects, "HCLine point not in objects"
        _lp.freeUser(hcl)
        del self.__objects[id(hcl)]
        self.__hclines.delObject(hcl)
        if all:
            self.delObject(_lp) # remove possibly unused point _lp

    def _delVCLine(self, vcl, all=True):
        """Remove a VCLine object from the Layer.

_delVCLine(vcl[, all=True])

This method is private to the Layer object.
        """
        _lp = vcl.getLocation()
        assert id(_lp) in self.__objects, "VCLine point not in objects"
        _lp.freeUser(vcl)
        del self.__objects[id(vcl)]
        self.__vclines.delObject(vcl)
        if all:
            self.delObject(_lp) # remove possibly unused point _lp

    def _delACLine(self, acl, all=True):
        """Remove an ACLine object from the Layer.

_delACLine(acl[, all=True])

This method is private to the Layer object.
        """
        _lp = acl.getLocation()
        assert id(_lp) in self.__objects, "ACLine point not in objects"
        _lp.freeUser(acl)
        del self.__objects[id(acl)]
        self.__aclines.delObject(acl)
        if all:
            self.delObject(_lp) # remove possibly unused point _lp

    def _delCLine(self, cl, all=True):
        """Delete a CLine from the Layer.

_delCLine(cl[, all=True])

This method is private to the Layer object.
        """
        _p1, _p2 = cl.getKeypoints()
        assert id(_p1) in self.__objects, "CLine point p1 not in objects"
        _p1.freeUser(cl)
        assert id(_p2) in self.__objects, "CLine point p2 not in objects"
        _p2.freeUser(cl)
        del self.__objects[id(cl)]
        self.__clines.delObject(cl)
        if all:
            self.delObject(_p1) # remove possibly unused point _p1
            self.delObject(_p2) # remove possibly unused point _p2

    def _delCCircle(self, cc, all=True):
        """Delete a CCircle from the Layer.

_delCCircle(cc[, all=True])

This method is private to the Layer object.
        """
        _cp = cc.getCenter()
        assert id(_cp) in self.__objects, "CCircle center point not in objects"
        _cp.freeUser(cc)
        del self.__objects[id(cc)]
        self.__ccircles.delObject(cc)
        if all:
            self.delObject(_cp) # remove possibly unused point _cp

    def _delChamfer(self, ch):
        """Remove a Chamfer from the Layer.

_delChamfer(ch)

This method is private to the Layer.
        """
        _chamfers = self.__chamfers
        _idx = None
        for _i in range(len(_chamfers)):
            if ch is _chamfers[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost chamfer from list"
        for _uref in ch.getUsers(): # could be hatching ...
            _user = _uref()
            if _user is not None:
                self.delObject(_user)
        _s1, _s2 = ch.getSegments()
        assert id(_s1) in self.__objects, "Chamfer s1 segment not in objects"
        _s1.freeUser(ch)
        assert id(_s2) in self.__objects, "Chamfer s2 segment not in objects"
        _s2.freeUser(ch)
        del self.__objects[id(ch)]
        del _chamfers[_idx] # restore the things the chamfer connects?

    def _delFillet(self, f):
        """Remove a Fillet from the Layer.

_delFillet(f)

This method is private to the Layer.
        """
        _fillets = self.__fillets
        _idx = None
        for _i in range(len(_fillets)):
            if f is _fillets[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost fillet from list"
        for _uref in f.getUsers(): # could be hatching ...
            _user = _uref()
            if _user is not None:
                self.delObject(_user)
        _s1, _s2 = f.getSegments()
        assert id(_s1) in self.__objects, "Fillet s1 segment not in objects"
        _s1.freeUser(f)
        assert id(_s2) in self.__objects, "Fillet s2 segment not in objects"
        _s2.freeUser(f)
        del self.__objects[id(f)]
        del _fillets[_idx] # restore the things the fillet connects?

    def _delLeader(self, l, all=True):
        """Delete a Leader from the Layer.

_delLeader(l[, all=True])

This method is private to the Layer object.
        """
        _p1, _p2, _p3 = l.getPoints()
        assert id(_p1) in self.__objects, "Leader p1 Point not in objects"
        _p1.freeUser(l)
        assert id(_p2) in self.__objects, "Leader p2 Point not in objects"
        _p2.freeUser(l)
        assert id(_p3) in self.__objects, "Leader p3 Point not in objects"
        _p3.freeUser(l)
        del self.__objects[id(l)]
        self.__leaders.delObject(l)
        if all:
            self.delObject(_p1) # remove possibly unused point _p1
            self.delObject(_p2) # remove possibly unused point _p2
            self.delObject(_p3) # remove possibly unused point _p3

    def _delPolyline(self, pl, all=True):
        """Delete a Polyline from the Layer.

_delPolyline(pl[, all=True])

This method is private to the Layer object.
        """
        for _uref in pl.getUsers(): # could be hatching
            _user = _uref()
            if _user is not None:
                self.delObject(_user)
        _pts = pl.getPoints()
        for _pt in _pts:
            assert id(_pt) in self.__objects, "Polyline point not in objects"
            _pt.freeUser(pl)
        del self.__objects[id(pl)]
        self.__polylines.delObject(pl)
        if all:
            for _pt in _pts:
                self.delObject(_pt) # remove possibly unused point _pt

    def _delTextBlock(self, tb):
        """Delete a TextBlock from the Layer.

_delTextBlock(tb)

This method is private to the Layer object.
        """
        _tbs = self.__textblocks
        _idx = None
        for _i in range(len(_tbs)):
            if tb is _tbs[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost textblock in list"
        del self.__objects[id(tb)]
        del _tbs[_idx]

    def _delAngularDimension(self, adim):
        """Delete an AngularDimension from the Layer.

_delAngularDimension(adim)

This method is private to the Layer object.
        """
        _l1, _l2, _l3 = adim.getDimLayers()
        _p1, _p2, _p3 = adim.getDimPoints()
        if _l1 is self:
            assert id(_p1) in self.__objects, "ADim p1 not in objects"
        _p1.freeUser(adim)
        if _l2 is self:
            assert id(_p2) in self.__objects, "ADim p2 not in objects"
        _p2.freeUser(adim)
        if _l3 is self:
            assert id(_p3) in self.__objects, "ADim p3 not in objects"
        _p3.freeUser(adim)
        del self.__objects[id(adim)]
        self.__adims.delObject(adim)

    def _delRadialDimension(self, rdim):
        """Delete a RadialDimension from the Layer.

_delRadialDimension(rdim)

This method is private to the Layer object.
        """
        _layer = rdim.getDimLayer()
        _circ = rdim.getDimCircle()
        if _layer is self:
            assert id(_circ) in self.__objects, "RDim circle not in objects"
        _circ.freeUser(rdim)
        del self.__objects[id(rdim)]
        self.__rdims.delObject(rdim)

    def _delHorizontalDimension(self, hdim):
        """Delete an HorizontalDimension from the Layer.

_delHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        _l1, _l2 = hdim.getDimLayers()
        _p1, _p2 = hdim.getDimPoints()
        if _l1 is self:
            assert id(_p1) in self.__objects, "HDim p1 not in objects"
        _p1.freeUser(hdim)
        if _l2 is self:
            assert id(_p2) in self.__objects, "HDim p2 not in objects"
        _p2.freeUser(hdim)
        del self.__objects[id(hdim)]
        self.__hdims.delObject(hdim)

    def _delVerticalDimension(self, vdim):
        """Delete an VerticalDimension from the Layer.

_delVerticalDimension(vdim)

This method is private to the Layer object.
        """
        _l1, _l2 = vdim.getDimLayers()
        _p1, _p2 = vdim.getDimPoints()
        if _l1 is self:
            assert id(_p1) in self.__objects, "VDim p1 not in objects"
        _p1.freeUser(vdim)
        if _l2 is self:
            assert id(_p2) in self.__objects, "VDim p2 not in objects"
        _p2.freeUser(vdim)
        del self.__objects[id(vdim)]
        self.__vdims.delObject(vdim)

    def _delLinearDimension(self, ldim):
        """Delete an LinearDimension from the Layer.

_delLinearDimension(ldim)

This method is private to the Layer object.
        """
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if _l1 is self:
            assert id(_p1) in self.__objects, "LDim p1 not in objects"
        _p1.freeUser(ldim)
        if _l2 is self:
            assert id(_p2) in self.__objects, "LDim p2 not in objects"
        _p2.freeUser(ldim)
        del self.__objects[id(ldim)]
        self.__ldims.delObject(ldim)

    def usesObject(self, obj):
        """Return a list of objects tied to another object

usesObject(obj)

The object "obj" should be a Point, Segment, Circle, Arc,
Chamfer, or Fillet.
        """
        warnings.warn("usesObject() is deprecated - test for Subpartness instead",
                      DeprecationWarning, stacklevel=2)
        _users = []
        if isinstance(obj, baseobject.Subpart):
            _users.extend(obj.getUsers())
        return _users

    def countUsers(self, obj):
        """Return the number of objects using a particular object.

countUsers(obj)

This method returns an integer value.
        """
        warnings.warn("countUsers() is deprecated - test for Subpartness instead",
                      DeprecationWarning, stacklevel=2)
        if id(obj) not in self.__objects:
            raise ValueError, "Object not stored in layer: " + `obj`
        _count = 0
        if isinstance(obj, baseobject.Subpart):
            _count = obj.countUsers()
        return _count

    def findObject(self, obj):
        """Return an object in the layer that is equivalent to a test object.

findObject(obj)

This method returns None if a suitable object is not found.
        """
        _retobj = None
        if id(obj) in self.__objects:
            _retobj = obj
        else:
            self.update()
            _objtree = None
            _objlist = None
            if isinstance(obj, point.Point):
                _x, _y = obj.getCoords()
                _retobj = self.__points.find(_x, _y)
            elif isinstance(obj, segment.Segment):
                _p1, _p2 = obj.getEndpoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _retobj = self.__segments.find(_x1, _y1, _x2, _y2)
            elif isinstance(obj, arc.Arc):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _sa = obj.getStartAngle()
                _ea = obj.getEndAngle()
                _retobj = self.__arcs.find(_x, _y, _r, _sa, _ea)
            elif isinstance(obj, circle.Circle):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _retobj = self.__circles.find(_x, _y, _r)
            elif isinstance(obj, hcline.HCLine):
                _y = obj.getLocation().y
                _retobj = self.__hclines.find(_y)
            elif isinstance(obj, vcline.VCLine):
                _x = obj.getLocation().x
                _retobj = self.__vclines.find(_x)
            elif isinstance(obj, acline.ACLine):
                _x, _y = obj.getLocation().getCoords()
                _angle = obj.getAngle()
                _retobj = self.__aclines.find(_x, _y, _angle)
            elif isinstance(obj, cline.CLine):
                _p1, _p2 = obj.getKeypoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _retobj = self.__clines.find(_x1, _y1, _x2, _y2)
            elif isinstance(obj, ccircle.CCircle):
                _x, _y = obj.getCenter().getCoords()
                _r = obj.getRadius()
                _retobj = self.__ccircles.find(_x, _y, _r)
            elif isinstance(obj, segjoint.Fillet):
                _objlist = self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                _objlist = self.__chamfers
            elif isinstance(obj, leader.Leader):
                _p1, _p2, _p3 = obj.getPoints()
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _x3, _y3 = _p3.getCoords()
                _retobj = self.__leaders.find(_x1, _y1, _x2, _y2, _x3, _y3)
            elif isinstance(obj, polyline.Polyline):
                _coords = []
                for _pt in obj.getPoints():
                    _coords.append(_pt.getCoords())
                _retobj = self.__polylines.find(_coords)
            elif isinstance(obj, text.TextBlock):
                _objlist = self.__textblocks
            elif isinstance(obj, dimension.HorizontalDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _retobj = self.__hdims.find(_l1, _p1, _l2, _p2)
            elif isinstance(obj, dimension.VerticalDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _retobj = self.__hdims.find(_l1, _p1, _l2, _p2)
            elif isinstance(obj, dimension.LinearDimension):
                _l1, _l2 = obj.getDimLayers()
                _p1, _p2 = obj.getDimPoints()
                _retobj = self.__hdims.find(_l1, _p1, _l2, _p2)
            elif isinstance(obj, dimension.RadialDimension):
                _l1 = obj.getDimLayer()
                _c1 = obj.getDimCircle()
                _retobj = self.__rdims.find(_l1, _c1)
            elif isinstance(obj, dimension.AngularDimension):
                _vl, _l1, _l2 = obj.getDimLayers()
                _vp, _p1, _p2 = obj.getDimPoints()
                _retobj = self.__hdims.find(_vl, _vp, _l1, _p1, _l2, _p2)
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
            if _objtree is not None:
                _retobj = _objtree.binscan(obj)
            elif _objlist is not None:
                _seen = True
                try:
                    _idx = _objlist.index(obj)
                except ValueError:
                    _seen = False
                if _seen:
                    _retobj = _objlist[_idx]
        return _retobj

    def find(self, typestr, *args):
        """Find an existing entity in the drawing.

find(typestr, *args)

typestr: A string giving the type of entity to find
*args: A variable number of arguments used for searching
        """
        if not isinstance(typestr, str):
            raise TypeError, "Invalid type string: " + str(typestr)
        _obj = None
        if typestr == 'point':
            _obj = self.__points.find(*args)
        elif typestr == 'segment':
            _obj = self.__segment.find(*args)
        elif typestr == 'circle':
            _obj = self.__circles.find(*args)
        elif typestr == 'arc':
            _obj = self.__arcs.find(*arg)
        elif typestr == 'hcline':
            _obj = self.__hclines.find(*args)
        elif typestr == 'vcline':
            _obj = self.__vclines.find(*args)
        elif typestr == 'acline':
            _obj = self.__aclines.fine(*args)
        elif typestr == 'cline':
            _obj = self.__clines.find(*args)
        elif typestr == 'ccircle':
            _obj = self.__ccircles.find(*args)
        elif typestr == 'leader':
            _obj = self.__leaders.find(*args)
        elif typestr == 'polyline':
            _obj = self.__polylines.find(*args)
        elif typestr == 'ldim':
            _obj = self.__ldims.find(*args)
        elif typestr == 'hdim':
            _obj = self.__hdims.find(*args)
        elif typestr == 'vdim':
            _obj = self.__vdims.find(*args)
        elif typestr == 'rdim':
            _obj = self.__rdims.find(*args)
        elif typestr == 'adim':
            _obj = self.__adims.find(*args)
        else:
            raise ValueError, "Unexpected type string '%s'" % typestr
        return _obj

    def hasPoint(self, p, tol=tolerance.TOL):
        """Find a Point in the layer.

hasPoint(p [,tol])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There is one optional argument:

tol: A float equal or greater than 0

The function returns an existing Point object found at
the same location, or None.
        """
        warnings.warn("hasPoint() is deprecated - use find()",
                      DeprecationWarning, stacklevel=2)
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _lp = None
        if id(_p) in self.__objects:
            _lp = _p
        else:
            _t = tolerance.toltest(tol)
            self.update()
            _x, _y = _p.getCoords()
            _lp = self.__points.find(_x, _y, _t)
        return _lp

    def hasPointAt(self, x, y):
        """Find a Point in the layer at a specific location.

hasPointAt(x, y)

The two required arguments are:

x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Point if one is found at
the location, or None otherwise.
        """
        warnings.warn("hasPointAt() is deprecated - use find()",
                      DeprecationWarning, stacklevel=2)
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__points.find(_x, _y)

    def hasCircleAt(self, r, x, y):
        """Find a Circle of some radius in the layer at a specific location.

hasCircleAt(r, x, y)

The three required arguments are:

r: A float giving the Circle radius
x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Circle if one is found at
the location, or None otherwise.
        """
        warnings.warn("hasCircleAt() is deprecated - use find()",
                      DeprecationWarning, stacklevel=2)
        _r = r
        if not isinstance(_r, float):
            _r = float(_r)
        if _r < 1e-10:
            raise ValueError, "Invalid radius: %g" % _r
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__circles.find(_x, _y, _r)

    def mapPoint(self, p, tol=tolerance.TOL, count=2):
        """Find a Point in the layer

mapPoint(p [,tol, count])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There are two optional arguments:

tol: A float equal or greater than 0 for distance tolerance comparisons.
count: An integer value indicating the largest number of objects to
       return. By default this value is 2.

Setting "count" to None or a negative value will result in the maximum
number of objects being unlimited.

This method tests the objects in the Layer to see if the
Point can can be mapped on to any of them. The returned list
consists of tuples in the form:

(obj, pt)

Where "obj" is the object the point was mapped to and "pt"
is the projected point on the object.
        """
        _scan = True
        _hits = []
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _count = count
        if _count is None:
            _count = sys.maxint
        else:
            if not isinstance(_count, int):
                _count = int(count)
            if _count < 0:
                _count = sys.maxint
        if _count < 1: # bail out, but why set count to this value?
            return _hits
        _x, _y = _p.getCoords()
        _xmin = _x - _t
        _xmax = _x + _t
        _ymin = _y - _t
        _ymax = _y + _t
        self.update()
        #
        # scan for non-point objects first, then look at points,
        # because there will not be any way the same object can
        # be found in differnt trees/lists
        #
        if _scan:
            for _seg in self.__segments.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _seg.getProjection(_x, _y)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_seg, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _circ in self.__circles.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _circ.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_circ, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _arc in self.__arcs.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _arc.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_arc, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _hcl in self.__hclines.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _hcl.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_hcl, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _vcl in self.__vclines.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _vcl.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_vcl, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _acl in self.__aclines.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _acl.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_acl, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _cc in self.__ccircles.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _cc.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_cc, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _cl in self.__clines.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _cl.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_cl, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _l in self.__leaders.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _l.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_l, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _pl in self.__polylines.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _pl.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_pl, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _ldim in self.__ldims.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _ldim.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_ldim, _lp))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _hdim in self.__hdims.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _hdim.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_hdim, _lp))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _vdim in self.__vdims.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _vdim.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_vdim, _lp))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _rdim in self.__rdims.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _rdim.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_rdim, _lp))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _adim in self.__adims.getInRegion(_xmin, _ymin, _xmax, _ymax):
                _pt = _adim.mapCoords(_x, _y, _t)
                if _pt is not None:
                    _px, _py = _pt
                    _lp = self.__points.find(_px, _py)
                    if _lp is None:
                        _lp = point.Point(_px, _py)
                    _hits.append((_adim, _lp))
                    if len(_hits) == _count:
                        break
        #
        # scan for point, but do not append any object that
        # has already been added to the hit list
        #
        if _scan:
            _objs = {}
            for _obj, _pt in _hits:
                _objs[id(_obj)] = True
            _lp = self.find('point', _x, _y, _t)
            if _lp is not None:
                assert id(_lp) in self.__objects, "Point not in objects"
                for _uref in _lp.getUsers():
                    _user = _uref()
                    if _user is not None:
                        _uid = id(_user)
                        if _uid not in _objs:
                            _objs[_uid] = True
                            _hits.append((_user, _lp))
                            if len(_hits) == _count:
                                _scan = False
                                break

        return _hits

    def getLayerEntities(self, entity):
        """Get all of a particular type of entity in the Layer.

getLayerEntities(entity)

The argument "entity" should be one of the following:
point, segment, circle, arc, hcline, vcline, acline,
cline, ccircle, chamfer, fillet, leader, polyline,
textblock, linear_dimension, horizontal_dimenions,
vertical_dimension, radial_dimension, or angular_dimension.
        """
        if not isinstance(entity, str):
            raise TypeError, "Invalid entity type %s" % type(entity)
        if entity == "point":
            _objs = self.__points.getObjects()
        elif entity == "segment":
            _objs = self.__segments.getObjects()
        elif entity == "circle":
            _objs = self.__circles.getObjects()
        elif entity == "arc":
            _objs = self.__arcs.getObjects()
        elif entity == "hcline":
            _objs = self.__hclines.getObjects()
        elif entity == "vcline":
            _objs = self.__vclines.getObjects()
        elif entity == "acline":
            _objs = self.__aclines.getObjects()
        elif entity == "cline":
            _objs = self.__clines.getObjects()
        elif entity == "ccircle":
            _objs = self.__ccircles.getObjects()
        elif entity == "chamfer":
            _objs = self.__chamfers[:]
        elif entity == "fillet":
            _objs = self.__fillets[:]
        elif entity == "leader":
            _objs = self.__leaders.getObjects()
        elif entity == "polyline":
            _objs = self.__polylines.getObjects()
        elif entity == "text":
            _objs = self.__textblocks[:]
        elif entity == "linear_dimension":
            _objs = self.__ldims.getObjects()
        elif entity == "horizontal_dimension":
            _objs = self.__hdims.getObjects()
        elif entity == "vertical_dimension":
            _objs = self.__vdims.getObjects()
        elif entity == "radial_dimension":
            _objs = self.__rdims.getObjects()
        elif entity == "angular_dimension":
            _objs = self.__adims.getObjects()
        else:
            raise ValueError, "Invalid layer entity '%s'" % entity
        return _objs

    def addChild(self, child):
        """Add a child Layer to a Layer.

addChild(child)

There is no limit to the number of children for a Layer.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        child.__parent = self
        self.__children.append(child)

    def delChild(self, child):
        """Remove a child Layer from the Layer.

delChild(child)

The child layer will not be removed if it has children
of its own.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        _seen = False
        _index = 0
        for _kid in self.__children:
            if child is _kid:
                _seen = True
                break
            _index = _index + 1
        if not _seen:
            raise ValueError, "Invalid child Layer: " + `child`
        if not child.hasChildren():
            del self.__children[_index]

    def hasChildren(self):
        """Test if there are children of this Layer.

hasChildren()

The function returns 1 if the Layer has children, 0 otherwise.
        """
        _val = False
        if len(self.__children) > 0:
            _val = True
        return _val

    def getChildren(self):
        """Get the children of a Layer.

getChildren()

This function returns a list of children Layers for a Layer.
        """
        return self.__children[:]

    def getParent(self):
        """Get the Layer's parent.

getParent()
        """
        return self.__parent

    def setParent(self, parent):
        """Set the Layer's parent.

setParent(parent)

Set the parent layer for a Layer object. The parent cannot
be set to None.
        """
        if parent is None:
            raise ValueError, "Layer parent cannot be set to None."
        if not isinstance(parent, Layer):
            raise TypeError, "Invalid parent layer: " + `parent`
        self.__parent = parent

    parent = property(getParent, setParent, None, "Parent of a layer.")

    def getVisibility(self):
        """Return the visibility value of the Layer.

getVisibility()
        """
        return self.__visibility

    def setVisibility(self, val):
        """Set the visibility of the Layer.

setVisibility(val)

The argument "val" should be either "True" or "False".
        """
        if val is not True and val is not False:
            raise ValueError, "Invalid visibility value: " + `val`
        self.__visibility = val

    visibility = property(getVisibility, setVisibility, None, "Visibility.")

    def getScale(self):
        """Return the scale factor of the Layer.

getScale()
        """
        return self.__scale

    def setScale(self, scale):
        """Set the scale factor for the Layer.

setScale(scale)

The scale factor must be a positive float value greater than 0.0
        """
        _s = scale
        if not isinstance(_s, float):
            _s = float(scale)
        if _s < 1e-10:
            raise ValueError, "Invalid scale factor: %g" % _s
        _os = self.__scale
        if abs(_os - _s) > 1e-10:
            self.__scale = _s
            self.sendMessage('attribute_changed', 'scale', _os)

    scale = property(getScale, setScale, None, "Layer scale factor.")

    def getBoundary(self):
        """Return the maximum and minimum values of the object in the Layer.

getBoundary()

The function returns a tuple holding four float values:

(xmin, ymin, xmax, _ymax)

A default value of (-1.0, -1.0, 1.0, 1.0) is returned for a Layer
containing no objects.
        """
        _xmin = None
        _ymin = None
        _xmax = None
        _ymax = None
        for _obj in self.__points.getObjects():
            _x, _y = _obj.getCoords()
            if _xmin is None or _x < _xmin:
                _xmin = _x
            if _ymin is None or _y < _ymin:
                _ymin = _y
            if _xmax is None or _x > _xmax:
                _xmax = _x
            if _ymax is None or _y > _ymax:
                _ymax = _y
        for _obj in self.__arcs.getObjects():
            _axmin, _aymin, _axmax, _aymax = _obj.getBounds()
            if _xmin is None or _axmin < _xmin:
                _xmin = _axmin
            if _ymin is None or _aymin < _ymin:
                _ymin = _aymin
            if _xmax is None or _axmax > _xmax:
                _xmax = _axmax
            if _ymax is None or _aymax > _ymax:
                _ymax = _aymax
        for _obj in self.__circles.getObjects() + self.__ccircles.getObjects():
            _x, _y = _obj.getCenter().getCoords()
            _r = _obj.getRadius()
            _val = _x - _r
            if _xmin is None or _val < _xmin:
                _xmin = _val
            _val = _y - _r
            if _ymin is None or _val < _ymin:
                _ymin = _val
            _val = _x + _r
            if _xmax is None or _val > _xmax:
                _xmax = _val
            _val = _y + _r
            if _ymax is None or _val > _ymax:
                _ymax = _val
        _dims = (self.__ldims.getObjects() +
                 self.__hdims.getObjects() +
                 self.__vdims.getObjects() +
                 self.__rdims.getObjects() +
                 self.__adims.getObjects())
        for _obj in _dims:
            _dxmin, _dymin, _dxmax, _dymax = _obj.getBounds()
            if _xmin is None or _dxmin < _xmin:
                _xmin = _dxmin
            if _ymin is None or _dymin < _ymin:
                _ymin = _dymin
            if _xmax is None or _dxmax > _xmax:
                _xmax = _dxmax
            if _ymax is None or _dymax > _ymax:
                _ymax = _dymax
        for _textblock in self.__textblocks:
            _x, _y = _textblock.getLocation() # need to get boundary of text
            if _xmin is None or _x < _xmin:
                _xmin = _x
            if _ymin is None or _y < _ymin:
                _ymin = _y
            if _xmax is None or _x > _xmax:
                _xmax = _x
            if _ymax is None or _y > _ymax:
                _ymax = _y
        if _xmin is None: _xmin = -1.0
        if _ymin is None: _ymin = -1.0
        if _xmax is None: _xmax = 1.0
        if _ymax is None: _ymax = 1.0
        return _xmin, _ymin, _xmax, _ymax

    def objsInRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return a all the objects in the Layer visible within the bounds.

objsInRegion(xmin, ymin, xmax, ymax[, fully])

The function has four required arguments:

xmin: The minimum x-value of the region
ymin: The minimum y-value of the region
xmax: The maximum x-value of the region
ymax: The maximum y-value of the region

There is a single optional argument:

fully: A True/False value indicating if the object must be
       entirely within the region [fully=True], or can
       merely pass through [fully=False]. The default value
       is False.

The function returns a list of objects.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if xmax < xmin:
            raise ValueError, "Value error: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Value error: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        self.update()
        _objs = []
        _objs.extend(self.__points.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__segments.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__circles.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__arcs.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__hclines.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__vclines.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__aclines.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__clines.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__ccircles.getInRegion(_xmin, _ymin, _xmax, _ymax))
        for _obj in self.__chamfers:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        _objs.extend(self.__leaders.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__polylines.getInRegion(_xmin, _ymin, _xmax, _ymax))
        for _obj in self.__fillets:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        _objs.extend(self.__ldims.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__hdims.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__vdims.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__rdims.getInRegion(_xmin, _ymin, _xmax, _ymax))
        _objs.extend(self.__adims.getInRegion(_xmin, _ymin, _xmax, _ymax))
        for _obj in self.__textblocks:
            _x, _y = _obj.getLocation()
            if _xmin < _x < _xmax and _ymin < _y < _ymax:
                _objs.append(_obj)
        return _objs

    def sendsMessage(self, m):
        if m in Layer.messages:
            return True
        return entity.Entity.sendsMessage(self, m)

    def update(self):
        """Check that the objects in this layer are stored correctly.

update()

This function checks that the objects held in this layer are kept
in the proper order. Also, any duplicated objects that may have
be created due to modifying entities in the layer are removed.
        """
        # if (self.__segment_tree.isModified() or
        # if (self.__circle_tree.isModified() or
        # if (self.__ccircle_tree.isModified() or
            # self.__hcline_tree.isModified() or
            # self.__vcline_tree.isModified() or
            # self.__acline_tree.isModified() or
            # self.__cline_tree.isModified() or
            # self.__ccircle_tree.isModified()):
            # self._sweepPointUsers()
        # self._sweepPointTree()
        # self._sweepSegmentTree()
        # self._sweepCircleTree()
        # self._sweepArcTree()
        # self._sweepSPCLineTree(self.__hcline_tree)
        # self._sweepSPCLineTree(self.__vcline_tree)
        # self._sweepSPCLineTree(self.__acline_tree)
        # self._sweepCLineTree()
        # self._sweepCCircleTree()
        # self._sweepLeaderList()
        # self._sweepPolylineList()

    def _sweepPointUsers(self):
        """Check that all objects associated with a point still do.

_sweepPointUsers()

This method is private to the Layer object.
        """
        # for _pt in self.__point_tree:
        for _pt in self.__points.getObjects():
            assert id(_pt) in self.__objects, "Lost point from objects"
            _freeobjs = []
            for _uref in _pt.getUsers():
                _user = _uref()
                if _user is not None:
                    if isinstance(_user, segment.Segment):
                        _p1, _p2 = _user.getEndpoints()
                        if _pt is not _p1 and _pt is not _p2:
                           _freeobjs.append(_user)
                    elif isinstance(_user, arc.Arc):
                        _cp = _user.getCenter()
                        _ep1, _ep2 = _user.getEndpoints()
                        if _pt is not _cp and _pt != _ep1 and _pt != _ep2:
                            _freeobjs.append(_user)
                    elif isinstance(_user, circle._Circle):
                        if _pt is not _user.getCenter():
                            _freeobjs.append(_user)
                    elif isinstance(_user, (hcline.HCLine,
                                            vcline.VCLine,
                                            acline.ACLine)):
                        _lp = _user.getLocation()
                        if _pt is not _lp:
                            _freeobjs.append(_user)
                    elif isinstance(_user, cline.CLine):
                        _p1, _p2 = _user.getKeypoints()
                        if _pt is not _p1 and _pt is not _p2:
                            _freeobjs.append(_user)
                    elif isinstance(_user, leader.Leader):
                        _p1, _p2, _p3 = _user.getPoints()
                        if ((_pt is not _p1) and
                            (_pt is not _p2) and
                            (_pt is not _p3)):
                            _freeobjs.append(_user)
                    elif isinstance(_user, polyline.Polyline):
                        if _pt not in _user.getPoints():
                            _freeobjs.append(_user)
                    elif isinstance(_user, dimension.Dimension):
                        if isinstance(_user, dimension.LinearDimension):
                            _p1, _p2 = _user.getDimPoints()
                            if _pt is not _p1 and _pt is not _p2:
                                _freeobjs.append(_user)
                        elif isinstance(_user, dimension.AngularDimension):
                            _p1, _p2, _p3 = _user.getDimPoints()
                            if (_pt is not _p1 and
                                _pt is not _p2 and
                                _pt is not _p3):
                                _freeobjs.append(_user)
            if len(_freeobjs):
                for _obj in _freeobjs:
                    print "Failed to free object from point: " + `_obj`
                    _pt.freeUser(_obj)

    def _sweepPointTree(self):
        """Validate the layer Point storage.

_sweepPointTree()

This method is private to the Layer object.
        """
        _modpts = self.__point_tree.getModified()
        if len(_modpts):
            _deadpts = self.__point_tree.clean()
            for _pt in _deadpts:
                _lp = self.__point_tree.binscan(_pt)
                assert _lp is not None, "lost point in point_tree: " + str(_pt)
                for _uref in _pt.getUsers():
                    _user = _uref()
                    if _user is not None:
                        if isinstance(_user, segment.Segment):
                            _p1, _p2 = _user.getEndpoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, circle._Circle):
                            _user.setCenter(_lp)
                        elif isinstance(_user, (hcline.HCLine,
                                                vcline.VCLine,
                                                acline.ACLine)):
                            _user.setLocation(_lp)
                        elif isinstance(_user, cline.CLine):
                            _p1, _p2 = _user.getKeypoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, leader.Leader):
                            _p1, _p2, _p3 = _user.getPoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            elif _pt is _p3:
                                _user.setP3(_lp)
                            else:
                                pass
                        elif isinstance(_user, polyline.Polyline):
                            for _i in range(len(_user)):
                                _pi = _user.getPoint(_i)
                                if _pi is _pt:
                                    _user.setPoint(_i, _lp)
                                    break
                        elif isinstance(_user, dimension.Dimension):
                            if isinstance(_user, dimension.LinearDimension):
                                _p1, _p2 = _user.getDimPoints()
                                if _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                            elif isinstance(_user, dimension.AngularDimension):
                                _vp, _p1, _p2 = _user.getDimPoints()
                                if _pt is _vp:
                                    _user.setVertexPoint(_lp)
                                elif _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                        else:
                            raise TypeError, "Missing type for keypoint modification:" + `_user`
            for _pt in _modpts:
                if _pt not in _deadpts:
                    for _uref in _pt.getUsers():
                        _user = _uref()
                        if _user is not None:
                            _user.modified()
                    _pt.reset()

    def _sweepSegmentTree(self):
        """Validate the layer segment storage.

_sweepSegmentTree()

This method is private to the Layer.
        """
        _modsegs = self.__segment_tree.getModified()
        if len(_modsegs):
            _deadsegs = self.__segment_tree.clean()
            for _seg in _modsegs:
                if _seg in _deadsegs:
                    continue
                if _seg.length() < 1e-10:
                    _deadsegs.append(_seg)
                    continue
                _seg.reset()
                for _pt in _seg.getEndpoints():
                    assert id(_pt) in self.__objects, "Segment point not in objects:" + `_pt`
            for _seg in _deadsegs:
                _keepseg = self.__segment_tree.binscan(_seg)
                assert _keepseg is not None, "Lost segment: " + str(_seg)
                _p1, _p2 = _seg.getEndpoints()
                _p1.freeUser(_seg)
                _p2.freeUser(_seg)
                for _uref in _seg.getUsers():
                    _user = _uref()
                    if _user is not None:
                        self.delObject(_user)
                del self.__objects[id(_seg)]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCircleTree(self):
        """Validate the layer Circle storage.

_sweepCircleTree(spctree)

This method is private to the Layer.
        """
        _modcircs = self.__circle_tree.getModified()
        if len(_modcircs):
            _deadcircs = self.__circle_tree.clean()
            for _circ in _modcircs:
                if _circ in _deadcircs:
                    continue
                if _circ.getRadius() < 1e-10: # this probably can't happen
                    _deadcircs.append(_circ)
                    continue
                _circ.reset()
                _cp = _circ.getCenter()
                assert id(_cp) in self.__objects, "Circle center not in objects:" + `_cp`
                for _uref in _circ.getUsers():
                    _user = _uref()
                    if _user is not None: # radial dimension or hatching
                        _user.modified()
            for _circ in _deadcircs:
                _keepcirc = self.__circle_tree.binscan(_circ)
                assert _keepcirc is not None, "Lost Circle: " + `_circ`
                for _uref in _circ.getUsers():
                    _user = _uref()
                    if _user is not None:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keepcirc)
                        else:
                            pass # fixme - hatching ...
                _cp = _circ.getCenter()
                _cp.freeUser(_circ)
                del self.__objects[id(_circ)]
                self.delObject(_cp)

    def _sweepArcTree(self):
        """Validate the layer Arc storage.

_sweepArcTree(spctree)

This method is private to the Layer.
        """
        _modarcs = self.__arc_tree.getModified()
        if len(_modarcs):
            _deadarcs = self.__arc_tree.clean()
            for _arc in _modarcs:
                if _arc in _deadarcs:
                    continue
                if _arc.getRadius() < 1e-10: # probably can't happen ...
                    _deadarcs.append(_arc)
                    continue
                _arc.reset()
                _cp = _arc.getCenter()
                assert id(_cp) in self.__objects, "Arc center not in objects:" + `_cp`
                for _ep in _arc.getEndpoints():
                    _lp = self.hasPoint(_ep)
                    if _lp is None:
                        print "Arc missing point: " + str(_arc)
                        print "Endpoint: " + str(_ep)
                        continue
                    assert id(_lp) in self.__objects, "Arc endpoint not in objects:" + `_lp`
                for _uref in _arc.getUsers():
                    _user = _uref()
                    if _user is not None: # radial/angular dimension or hatching
                        _user.modified()
            for _arc in _deadarcs:
                _keeparc = self.__arc_tree.binscan(_arc)
                assert _keeparc is not None, "Lost Arc: " + `_arc`
                for _uref in _arc.getUsers():
                    _user = _uref()
                    if _user is not None:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keeparc)
                        elif isinstance(_user, dimension.AngularDimension):
                            pass # fixme
                        else:
                            pass # fixme - hatching ...
                _cp = _arc.getCenter()
                _cp.freeUser(_arc)
                for _ep in _arc.getEndpoints():
                    _p = self.hasPoint(_ep)
                    _p.freeUser(_arc)
                    self.delObject(_p)
                del self.__objects[id(_arc)]
                self.delObject(_cp)

    def _sweepSPCLineTree(self, spctree):
        """Validate the storage of single point construction line objects.

_sweepSPCLineTree(spctree)

This method is private to the Layer.
        """
        _modspclines = spctree.getModified()
        if len(_modspclines):
            _deadspclines = spctree.clean()
            for _spcline in _modspclines:
                if _spcline not in _deadspclines:
                    _spcline.reset()
                    _pt =  _spcline.getLocation()
                    assert id(_pt) in self.__objects, "spcline point not in objects:" + str(_spcline)
            for _spcline in _deadspclines:
                _pt = _spcline.getLocation()
                _pt.freeUser(_spcline)
                del self.__objects[id(_spcline)]
                self.delObject(_pt)

    def _sweepCLineTree(self):
        """Validate the layer CLine storage.

_sweepCLineTree()

This method is private to the Layer.
        """
        _modclines = self.__cline_tree.getModified()
        if len(_modclines):
            _deadclines = self.__cline_tree.clean()
            for _cline in _modclines:
                if _cline not in _deadclines:
                    _cline.reset()
                    for _pt in _cline.getKeypoints():
                        assert id(_pt) in self.__objects, "CLine keypoint not in objects:" + `_pt`
            for _cline in _deadclines:
                _p1, _p2 = _cline.getKeypoints()
                _p1.freeUser(_cline)
                _p2.freeUser(_cline)
                del self.__objects[id(_cline)]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCCircleTree(self):
        """Validate the layer CCircle storage.

_sweepCCircleTree()

This method is private to the Layer object.
        """
        _modccs = self.__ccircle_tree.getModified()
        if len(_modccs):
            _deadccs = self.__ccircle_tree.clean()
            for _cc in _modccs:
                if _cc in _deadccs:
                    continue
                if _cc.getRadius() < 1e-10: # probably not possible
                    _deadccs.append(_cc)
                else:
                    _cc.reset()
                    _cp = _cc.getCenter()
                    assert id(_cp) in self.__objects, "CCircle center not in objjects:" + `_cp`
            for _cc in _deadccs:
                _cp = _cc.getCenter()
                _cp.freeUser(_cc)
                del self.__objects[id(_cc)]
                self.delObject(_cp)

    def _sweepLeaderList(self):
        """Validate the layer Leader storage

_sweepLeaderList()

This method is private to the Layer.
        """
        _mods = []
        for _leader in self.__leaders:
            if _leader.isModified():
                _mods.append(_leader)
        if len(_mods):
            self._sweepPointUsers()
            for _leader in _mods:
                _leader.reset()
                _p1, _p2, _p3 = _leader.getPoints()
                assert id(_p1) in self.__objects, "Leader P1 point not in objects."
                assert id(_p2) in self.__objects, "Leader P2 point not in objects."
                assert id(_p3) in self.__objects, "Leader P3 point not in objects."
    def _sweepPolylineList(self):
        """Validate the layer Polyline storage

_sweepPolylineList()

This method is private to the Layer.
        """
        _mods = []
        for _polyline in self.__polylines:
            if _polyline.isModified():
                _mods.append(_polyline)
        if len(_mods):
            self._sweepPointUsers()
            for _polyline in _mods:
                _polyline.reset()
                for _pt in _polyline.getPoints():
                    assert id(_pt) in self.__objects, "Polyline point not in objects."
