#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# classes for points
#

import math
import tolerance
import util
import tree
import baseobject
import quadtree

class Point(baseobject.Subpart):
    """A 2-D point Class.

A Point has the following attributes:

x: x-coordinate
y: y-coordinate

A Point object has the following methods:

{get/set}x(): Get/Set the x-coordinate of the Point.
{get/set}y(): Get/Set the y-coordinate of the Point.
getCoords(): Return the Point x and y coordinates in a tuple.
move(): Move a Point.
clone(): Return an identical copy of a Point.
inRegion(): Returns True if the point is in some area.
    """
    messages = {
        'moved' : True,
    }
    # member functions

    def __init__(self, x, y=None):
        """Initialize a Point.

There are two ways to initialize a Point:

Point(xc,yc) - Two arguments, with both arguments being floats
Point((xc,yc)) - A single tuple containing two float objects
        """
        baseobject.Subpart.__init__(self)
        if isinstance(x, tuple):
            if y is not None:
                raise SyntaxError, "Invalid call to Point()"
            _x, _y = util.tuple_to_two_floats(x)
        elif y is not None:
            _x = x
            if not isinstance(_x, float):
                _x = float(x)
            _y = y
            if not isinstance(_y, float):
                _y = float(y)
        else:
            raise SyntaxError, "Invalid call to Point()."
        self.__x = _x
        self.__y = _y

    def __str__(self):
        return "(%g,%g)" % (self.__x, self.__y)

    def __sub__(self, p):
        """Return the separation between two points.

This function permits the use of '-' to be an easy to read
way to find the distance between two Point objects.
        """
        if not isinstance(p, Point):
            raise TypeError, "Invalid type for Point subtraction: " + str(p)
        _px, _py = p.getCoords()
        return math.hypot((self.__x - _px), (self.__y - _py))

    def __eq__(self, obj):
        """Compare a Point to either another Point or a tuple for equality.
        """
        if not isinstance(obj, (Point,tuple)):
            return False
        if isinstance(obj, Point):
            if obj is self:
                return True
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        if abs(self.__x - _x) < 1e-10 and abs(self.__y - _y) < 1e-10:
            return True
        return False

    def __ne__(self, obj):
        """Compare a Point to either another Point or a tuple for inequality.
        """
        if not isinstance(obj, (Point,tuple)):
            return True
        if isinstance(obj, Point):
            if obj is self:
                return False
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        if abs(self.__x - _x) < 1e-10 and abs(self.__y - _y) < 1e-10:
            return False
        return True

    def getx(self):
        """Return the x-coordinate of a Point.

getx()
        """
        return self.__x

    def setx(self, val):
        """Set the x-coordinate of a Point

setx(val)

The argument "val" must be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Coordinate change not allowed - object locked."
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        _x = self.__x
        if abs(_x - _v) > 1e-10:
            self.__x = _v
            self.sendMessage('moved', _x, self.__y)
            self.modified()

    x = property(getx, setx, None, "x-coordinate value")

    def gety(self):
        """Return the y-coordinate of a Point.

gety()
        """
        return self.__y

    def sety(self, val):
        """Set the y-coordinate of a Point

sety(val)

The argument "val" must be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Coordinate change not allowed - object locked."
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        _y = self.__y
        if abs(_y - _v) > 1e-10:
            self.__y = _v
            self.sendMessage('moved', self.__x, _y)
            self.modified()

    y = property(gety, sety, None, "y-coordinate value")

    def getCoords(self):
        """Return the x and y Point coordinates in a tuple.

getCoords()
        """
        return (self.__x, self.__y)

    def move(self, dx, dy):
        """Move a Point.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Moving not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if not self.isLocked():
            if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
                _x = self.__x
                _y = self.__y
                self.__x = _x + _dx
                self.__y = _y + _dy
                self.sendMessage('moved', _x, _y)
                self.modified()

    def clone(self):
        """Create an identical copy of a Point.

clone()
        """
        return Point(self.__x, self.__y)

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Returns True if the Point is within the bounding values.

inRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the Point lies within that area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        return _xmin < self.__x < _xmax and _ymin < self.__y < _ymax

    def sendsMessage(self, m):
        if m in Point.messages:
            return True
        return baseobject.Subpart.sendsMessage(self, m)

class _Point(object):
    """A 2-D point Class.

A _Point has the following attributes:

x: x-coordinate
y: y-coordinate

A _Point object has the following methods:

{get/set}x(): Get/Set the x-coordinate of the _Point.
{get/set}y(): Get/Set the y-coordinate of the _Point.
getCoords(): Return the _Point x and y coordinates in a tuple.
move(): Move a _Point.
clone(): Return an identical copy of a _Point.
inRegion(): Returns True if the point is in some area.
    """

    # member functions

    def __init__(self, x, y=None):
        """Initialize a _Point.

There are two ways to initialize a _Point:

_Point(xc,yc) - Two arguments, with both arguments being floats
_Point((xc,yc)) - A single tuple containing two float objects
        """
        if isinstance(x, tuple):
            if y is not None:
                raise SyntaxError, "Invalid call to _Point()"
            _x, _y = util.tuple_to_two_floats(x)
        elif y is not None:
            _x = x
            if not isinstance(_x, float):
                _x = float(x)
            _y = y
            if not isinstance(_y, float):
                _y = float(y)
        else:
            raise SyntaxError, "Invalid call to _Point()."
        self.__x = _x
        self.__y = _y

    def __str__(self):
        return "(%g,%g)" % (self.__x, self.__y)

    def __sub__(self, p):
        """Return the separation between two points.

This function permits the use of '-' to be an easy to read
way to find the distance between two _Point objects.
        """
        if not isinstance(p, _Point):
            raise TypeError, "Invalid type for _Point subtraction: " + str(p)
        _px, _py = p.getCoords()
        return math.hypot((self.__x - _px), (self.__y - _py))

    def __zeq__(self, obj):
        """Compare a _Point to either another _Point or a tuple for equality.
        """
        if not isinstance(obj, (_Point,tuple)):
            raise TypeError, "Invalid comparison to %s" % type(obj)
        if isinstance(obj, _Point):
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        _val = False
        if abs(self.__x - _x) < 1e-10:
            if abs(self.__y - _y) < 1e-10:
                _val = True
        return _val

    def __zne__(self, obj):
        """Compare a _Point to either another _Point or a tuple for inequality.
        """
        if not isinstance(obj, (_Point,tuple)):
            raise TypeError, "Invalid comparison to %s" % type(obj)
        if isinstance(obj, _Point):
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        _val = True
        if abs(self.__x - _x) < 1e-10:
            if abs(self.__y - _y) < 1e-10:
                _val = False
        return _val

    def __cmp__(self, obj):
        """Compare a _Point to either another point or a two-float tuple.

The comparison of a _Point goes as follows:

1) If comparing to a tuple, get the two values from the
   tuple and treat them like "x" and "y" coordinates
2) If comparing to a _Point, get the _Point "x" and "y" attributes

The return value is determined as follows:

"x" value of P1 < "x" value of P2 ==> return -1
"x" value of P1 > "x" value of P2 ==> return 1

If we get here, the "y" values are compared

"y" value of P1 < "y" value of P2 ==> return -1
"y" value of P1 > "y" value of P2 ==> return 1

If we get here, return 0
        """
        if not isinstance(obj, (_Point,tuple)):
            raise TypeError, "Invalid type for comparison: " + `obj`
        if isinstance(obj, _Point):
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        _xs = self.__x
        _ys = self.__y
        if abs(_xs - _x) < 1e-10:
            if abs(_ys - _y) < 1e-10:
                _val = 0
            elif _ys > _y:
                _val = 1
            else:
                _val = -1
        elif _xs > _x:
            _val = 1
        else:
            _val = -1
        return _val

    def getx(self):
        """Return the x-coordinate of a _Point.

getx()
        """
        return self.__x

    def setx(self, val):
        """Set the x-coordinate of a _Point

setx(val)

The argument "val" must be a float.
        """
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        self.__x = _v

    x = property(getx, setx, None, "x-coordinate value")

    def gety(self):
        """Return the y-coordinate of a _Point.

gety()
        """
        return self.__y

    def sety(self, val):
        """Set the y-coordinate of a _Point

sety(val)

The argument "val" must be a float.
        """
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        self.__y = _v

    y = property(gety, sety, None, "y-coordinate value")

    def getCoords(self):
        """Return the x and y _Point coordinates in a tuple.

getCoords()
        """
        return (self.__x, self.__y)

    def move(self, dx, dy):
        """Move a _Point.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__x = self.__x + _dx
        self.__y = self.__y + _dy

    def clone(self):
        """Create an identical copy of a _Point.

clone()
        """
        return _Point(self.__x, self.__y)

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Returns True if the _Point is within the bounding values.

inRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the _Point lies within that area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        return _xmin < self.__x < _xmax and _ymin < self.__y < _ymax

class old_Point(baseobject.Subpart, tree.TreeItem, _Point):
    """An extension of _Point objects that can be stored in Tree objects.

The Point class is a wrapper around the _Point base class and
the TreeItem base class. This class extends several methods
in the base _Point class.
    """
    def __init__(self, x, y=None):
        """Initialize a Point object.

Refer to the  _Point.__init__() doc string for more info.
        """
        baseobject.Subpart.__init__(self)
        tree.TreeItem.__init__(self)
        _Point.__init__(self, x, y)

    def setx(self, x):
        """Set the x-coordinate of a Point.

setx(x)

This method extends _Point::setx() and sets the
modified flag indicating that the Point has been
altered.
        """
        _Point.setx(self, x)
        self.modified()

    x = property(_Point.getx, setx, None, "X Coordinate")

    def sety(self, y):
        """Set the y-coordinate of a Point.

sety(y)

This method extends _Point::sety() and sets the
modified flag indicating that the Point has been
altered.
        """
        _Point.sety(self, y)
        self.modified()

    y = property(_Point.gety, sety, None, "Y Coordinate")

    def move(self, dx, dy):
        """Move a Point.

move(dx, dy)

This method extends _Point::move() and sets the
modified flag indicating that the Point has been
altered.
        """
        _Point.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Point.

clone()

This method overrides the _Point::clone() method.
        """
        _x, _y = self.getCoords()
        return Point(_x, _y)

class PointTree(tree.Tree):
    """A Tree class specific for Point objects.

This class is based on the Tree class, and only
holds Point objects.
    """
    def __init__(self):
        """Initialize a PointTree.

PointTree()
        """
        tree.Tree.__init__(self, Point)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Point held in a PointTree via a binary-search

binscan(obj [, tol])

This method has one required parameter:

obj: Either a Point object or a tuple with exactly two floats

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Point object is found in the PointTree within the tolerance,
the index of that object is returned. Otherwise,the function returns
None.
        """
        _o = obj
        if isinstance(_o, Point):
            _x, _y = _o.getCoords()
        elif isinstance(_o, tuple):
            _x, _y = util.tuple_to_two_floats(_o)
        else:
            raise TypeError, "Invalid type for PointTree searching: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "PointTree in modified state."
        _scanlist = []
        _lo = _mid = 0
        _hi = len(self)
        _ro = None
        _xmin = _x - _t - 1e-10
        while _lo < _hi: # find the offset just before p.x > xmin ...
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _pt = self[_mid]
            if _pt.x > _xmin:
                _hi = _mid
            else:
                _lo = _mid + 1
        _ymin = _y - _t - 1e-10
        _xmax = _x + _t + 1e-10
        _ymax = _y + _t + 1e-10
        for _i in range(_mid, len(self)): # ...check points up to p.x > xmax
            _sobj = self[_i]
            _sx = _sobj.getx()
            if _sx > _xmax:
                break
            if _xmin < _sx and _ymin < _sobj.y < _ymax:
                _ro = _sobj
                break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a Point held in a PointTree via a linear-search

scan(obj [, tol])

This method has one required parameter:

obj: Either a Point object or a tuple with exactly two floats

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Point object is found in the PointTree within the tolerance,
the index of that object is returned. Otherwise,the function returns
None.
        """
        _o = obj
        if isinstance(_o, Point):
            _x, _y = _o.getCoords()
        elif isinstance(_o, tuple):
            _x, _y = util.tuple_to_two_floats(_o)
        else:
            raise TypeError, "Invalid type for PointTree searching: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "PointTree in modified state."
        _xmin = _x - _t - 1e-10
        _ymin = _y - _t - 1e-10
        _xmax = _x + _t + 1e-10
        _ymax = _y + _t + 1e-10
        _ro = None
        for _sobj in self:
            _sx = _sobj.getx()
            if _sx > _xmax:
                break
            if _xmin < _sx and _ymin < _sobj.y < _ymax:
                _ro = _sobj
                break
        return _ro

    def atCoords(self, x, y):
        """Find a Point in the PointTree at a certain location.

atCoords(x, y)

This method has two required arguments:

x: A float value representing the "x" coordinate
y: A float value representing the "y" coordinate

If there is a Point in the PointTree at the designated
coordinates, the Point is returned. Otherwise this method
returns None.

This method is similar to the binscan() method.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        if self.isModified():
            raise ValueError, "PointTree in modified state."
        _scanlist = []
        _lo = _mid = 0
        _hi = len(self)
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _pt = self[_mid]
            _px, _py = _pt.getCoords()
            if abs(_px - _x) < 1e-10:
                if abs(_py - _y) < 1e-10:
                    _ro = _pt
                    break
                elif _py < _y:
                    _lo = _mid + 1
                else:
                    _hi = _mid
            elif _px < _x:
                _lo = _mid + 1
            else:
                _hi = _mid
        return _ro

#
# Quadtree Point storage
#

class PointQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, Point):
            raise TypeError, "Invalid Point object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        while True:
            if ((_x > _xmax) or
                (_y > _ymax) or
                (_x < _xmin) or
                (_y < _ymin)):
                print "invalid node:"
                print "boundary: (%g, %g) -> (%g, %g)" % (_xmin, _ymin,
                                                          _xmax, _ymax)
                print "point: (%g, %g)" % (_x, _y)
                raise ValueError, "Failed to add point: " + `obj`
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.addObject(obj)
                obj.connect('moved', self, PointQuadtree.movePoint)
                break
        quadtree.Quadtree.addObject(self, obj)

    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCoords()
        _node = self.getTreeRoot()
        while True:
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x > _xmax) or
                (_y > _ymax) or
                (_x < _xmin) or
                (_y < _ymin)):
                print "invalid node:"
                print "boundary: (%g, %g) -> (%g, %g)" % (_xmin, _ymin,
                                                          _xmax, _ymax)
                print "point: (%g, %g)" % (_x, _y)
                raise ValueError, "Failed to delete point: " + `obj`
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.delObject(obj)
                _parent = _node.getParent()
                if _parent is not None:
                    self.purgeSubnodes(_parent)
                break
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)

    def find(self, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _t = tolerance.TOL
        if _alen > 2 :
            _t = tolerance.toltest(args[2])
        if not len(self):
            return None
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _p = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_y < (_ymin - _t)) or
                (_x > (_xmax + _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _xmin, _ymin, _xmax, _ymax = _subnode.getBoundary()
                    if ((_xmin - _t) < _x < (_xmax + _t) and
                        (_ymin - _t) < _y < (_ymax + _t)):
                        _nodes.append(_subnode)
            else:
                for _pt in _node.getObjects():
                    _xp, _yp = _pt.getCoords()
                    _xsep = _xp - _x
                    _ysep = _yp - _y
                    if abs(_xsep) < _t and abs(_ysep) < _t:
                        if _p is None:
                            _p = _pt
                        else:
                            if math.hypot(_xsep, _ysep) < (_p - _pt):
                                _p = _pt
        return _p

    def movePoint(self, obj, *args):
        if obj not in self:
            raise ValueError, "Point not stored in Quadtree: " + `obj`
        _alen = len(args)
        if len(args) < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _node = self.getTreeRoot()
        _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
        while True:
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.delObject(obj)
                quadtree.Quadtree.delObject(self, obj)
                obj.disconnect(self)
                break
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        return self.find(x, y, tol)

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _pts = []        
        if not len(self):
            return _pts
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _pt in _node.getObjects():
                    if _pt.inRegion(_xmin, _ymin, _xmax, _ymax):
                        _pts.append(_pt)
        return _pts
