#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for line segments
#
# TODO: Look at using weak references

import math
import warnings

import graphicobject
import baseobject
import tolerance
import style
import linetype
import color
import point
import tree
import maptree
import util
import quadtree

class Segment(graphicobject.new_GraphicObject):
    """A class representing a line segment.

A Segment has two attributes.

p1: A Point object representing the first end point
p2: A Point object representing the second end point

A Segment has the following methods:

getEndpoints(): Return the two endpoints of the Segment.
{get/set}P1: Get/Set the Segment first endpoint.
{get/set}P2: Get/Set the Segment second endpoint.
move(): Move the Segment.
length(): Get the Segment length.
getCoefficients(): Return the segment as ax + by + c = 0
mapPoint(): Test if a point is within some distance to a Segment.
getProjection(): Project a coordinate on to the Segment
mapCoords(): Test if a coordinate pair is within some distance to a Segment.
inRegion(): Test if the Segment is visible in some area.
clone(): Make an identical copy of a Segment.
    """
    messages = {
        'moved' : True,
        'endpoint_changed' : True
        }

    __defstyle = style.Style(u'Default Segment Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, p1, p2, st=None, lt=None, col=None, th=None):
        """Initialize a Segment object.

Segment(p1, p2[, st, lt, col, t])

p1: Segment first endpoint - may be a Point or a two-item tuple of floats.
p2: Segment second endpoint - may be a Point or a two-item tuple of floats.

The following arguments are optional:

st: A Style object
lt: A Linetype object that overrides the linetype in the Style.
col: A Color object that overrides the color in the Style.
t: A float that overrides the line thickness in the Style.
        """

        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)

        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 is _p2:
            raise ValueError, "Segments cannot have identical endpoints."
        _st = st
        if _st is None:
            _st = Segment.__defstyle
        graphicobject.new_GraphicObject.__init__(self, _st, lt, col, th)
        self.__p1 = _p1
        _p1.connect('moved', self, Segment.movePoint)
        _p1.storeUser(self)
        self.__p2 = _p2
        _p2.connect('moved', self, Segment.movePoint)
        _p2.storeUser(self)

    def __str__(self):
        return "Segment: %s to %s" % (self.__p1, self.__p2)

    def __eq__(self, obj):
        """Compare a Segment to another for equality.
        """
        if not isinstance(obj, Segment):
            return False
        if obj is self:
            return True
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = True
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = True
        else:
            _val = False
        return _val

    def __ne__(self, obj):
        """Compare a Segment to another for inequality.
        """
        if not isinstance(obj, Segment):
            return True
        if obj is self:
            return False
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = False
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = False
        else:
            _val = True
        return _val

    def getEndpoints(self):
        """Get the endpoints of the Segment.

getEndpoints()

This function returns a tuple containing the two Point objects
that are the endpoints of the segment.
        """
        return self.__p1, self.__p2

    def getP1(self):
        """Return the first endpoint Point of the Segment.

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint Point of the Segment.

setP1(p)
        """
        if self.isLocked():
            raise RuntimeError, "Setting endpoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid P1 endpoint: " + str(p)
        if p is self.__p2:
            raise ValueError, "Segments cannot have identical endpoints."
        _pt = self.__p1
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p1 = p
            self.sendMessage('endpoint_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Segment.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                _x, _y = self.__p2.getCoords()
                self.sendMessage('moved', _pt.x, _pt.y, _x, _y)
            self.modified()

    p1 = property(getP1, setP1, None, "First endpoint of the Segment.")

    def getP2(self):
        """Return the second endpoint Point of the Segment.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the second endpoint Point of the Segment.

setP2(p)
        """
        if self.isLocked():
            raise RuntimeError, "Setting endpoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + str(p)
        if p is self.__p1:
            raise ValueError, "Segments cannot have identical endpoints."
        _pt = self.__p2
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p2 = p
            self.sendMessage('endpoint_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Segment.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                _x, _y = self.__p1.getCoords()
                self.sendMessage('moved', _x, _y, _pt.x, _pt.y)
            self.modified()

    p2 = property(getP2, setP2, None, "Second endpoint of the Segment.")

    def move(self, dx, dy):
        """Move a Segment.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked() or self.__p1.isLocked() or self.__p2.isLocked():
            raise RuntimeError, "Moving Segment not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x1, _y1 = self.__p1.getCoords()
            _x2, _y2 = self.__p2.getCoords()
            self.mute()
            self.__p1.move(_dx, _dy)
            self.__p2.move(_dx, _dy)
            self.unmute()
            self.sendMessage('moved', _x1, _y1, _x2, _y2)
            self.modified()

    def length(self):
        """Return the length of the Segment.

length()
        """
        return self.__p1 - self.__p2

    def getCoefficients(self):
        """Express the line segment as a function ax + by + c = 0

getCoefficients()

This method returns a tuple of three floats: (a, b, c)
        """
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _a = _y2 - _y1
        _b = _x1 - _x2
        _c = (_x2 * _y1) - (_x1 * _y2)
        return _a, _b, _c

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the Segment. If the distance between the actual
Point and the Point used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        warnings.warn("mapPoint() is deprecated - use mapCoords()",
                      DeprecationWarning, stacklevel=2)
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def getProjection(self, x, y):
        """Find the projection point of some coordinates on the Segment.

getProjection(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _sqlen = pow((_x2 - _x1), 2) +  pow((_y2 - _y1), 2)
        if _sqlen < 1e-10: # coincident points
            return None
        _rn = ((_x - _x1) * (_x2 - _x1)) + ((_y - _y1) * (_y2 - _y1))
        _r = _rn/_sqlen
        if _r < 0.0 or _r > 1.0:
            return None
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        return _px, _py

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by coordinate pair to an
actual Point on the Segment. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        return util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a Segment exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the Segment lies within that area. If
the optional argument fully is used and is True, then both
endpoints of the Segment must lie within the boundary.
Otherwise, the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + str(fully)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _pxmin = min(_x1, _x2)
        _pymin = min(_y1, _y2)
        _pxmax = max(_x1, _x2)
        _pymax = max(_y1, _y2)
        if ((_pxmax < _xmin) or
            (_pymax < _ymin) or
            (_pxmin > _xmax) or
            (_pymin > _ymax)):
            return False
        if fully:
            if ((_pxmin > _xmin) and
                (_pymin > _ymin) and
                (_pxmax < _xmax) and
                (_pymax < _ymax)):
                return True
            return False
        return util.in_region(_x1, _y1, _x2, _y2, _xmin, _ymin, _xmax, _ymax)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        """Clip the Segment using the Liang-Barsky Algorithm.

clipToRegion(xmin, ymin, xmax, ymax)
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        #
        # simple tests to reject line
        #
        if ((_x1 < _xmin and _x2 < _xmin) or
            (_x1 > _xmax and _x2 > _xmax) or
            (_y1 < _ymin and _y2 < _ymin) or
            (_y1 > _ymax and _y2 > _ymax)):
            return None
        #
        # simple tests to accept line
        #
        _coords = None
        if (_xmin < _x1 < _xmax and
            _xmin < _x2 < _xmax and
            _ymin < _y1 < _ymax and
            _ymin < _y2 < _ymax):
            _coords = (_x1, _y1, _x2, _y2)
        else:
            #
            # the Segment can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # The following is the Liang-Barsky Algorithm
            # for segment clipping
            #
            _dx = _x2 - _x1
            _dy = _y2 - _y1
            _P = [-_dx, _dx, -_dy, _dy]
            _q = [(_x1 - _xmin), (_xmax - _x1), (_y1 - _ymin), (_ymax - _y1)]
            _u1 = 0.0
            _u2 = 1.0
            _valid = True
            for _i in range(4):
                _pi = _P[_i]
                _qi = _q[_i]
                if abs(_pi) < 1e-10:
                    if _qi < 0.0:
                        _valid = False
                        break
                else:
                    _r = _qi/_pi
                    if _pi < 0.0:
                        if _r > _u2:
                            _valid = False
                            break
                        if _r > _u1:
                            _u1 = _r
                    else:
                        if _r < _u1:
                            _valid = False
                            break
                        if _r < _u2:
                            _u2 = _r
            if _valid:
                _coords = (((_u1 * _dx) + _x1),
                           ((_u1 * _dy) + _y1),
                           ((_u2 * _dx) + _x1),
                           ((_u2 * _dy) + _y1))
        return _coords

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _p1 = self.__p1
        _p2 = self.__p2
        if p is _p1:
            _x1 = _x
            _y1 = _y
            _x2, _y2 = _p2.getCoords()
        elif p is _p2:
            _x1, _y1 = _p1.getCoords()
            _x2 = _x
            _y2 = _y
        else:
            raise ValueError, "Unexpected Segment endpoint: " + `p`
        self.sendMessage('moved', _x1, _y1, _x2, _y2)
        self.modified()

    def clone(self):
        """Create an identical copy of a Segment.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Segment(_cp1, _cp2, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Segment.messages:
            return True
        return graphicobject.new_GraphicObject.sendsMessage(self, m)

class _Segment(object):
    """A class representing a line segment.

A _Segment has two attributes.

p1: A _Point object representing the first end point
p2: A _Point object representing the second end point

A _Segment has the following methods:

getEndpoints(): Return the two endpoints of the _Segment.
{get/set}P1: Get/Set the _Segment first endpoint.
{get/set}P2: Get/Set the _Segment second endpoint.
move(): Move the _Segment.
length(): Get the _Segment length.
getCoefficients(): Return the segment as ax + by + c = 0
mapPoint(): Test if a point is within some distance to a _Segment.
getProjection(): Project a coordinate on to the Segment
mapCoords(): Test if a coordinate pair is within some distance to a _Segment.
inRegion(): Test if the _Segment is visible in some area.
clone(): Make an identical copy of a _Segment.
    """

    # member functions

    def __init__(self, p1, p2):
        """Instantiate a _Segment object.

_Segment(p1, p2)

p1: _Segment first endpoint - may be a _Point or a two-item tuple of floats
p2: _Segment second endpoint - may be a _Point or a two-item tuple of floats
        """
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 == _p2:
            raise ValueError, "_Segment objects cannot have equivalent endpoints."
        self.__p1 = _p1
        self.__p2 = _p2

    def __str__(self):
        return "Segment: %s to %s" % (self.__p1, self.__p2)

    def __zeq__(self, obj):
        """Compare a _Segment to another for equality.
        """
        if not isinstance(obj, _Segment):
            raise TypeError, "Invalid comparision to %s" % type(obj)
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = True
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = True
        else:
            _val = False
        return _val

    def __zne__(self, obj):
        """Compare a _Segment to another for inequality.
        """
        if not isinstance(obj, _Segment):
            raise TypeError, "Invalid comparision to %s" % type(obj)
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = False
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = False
        else:
            _val = True
        return _val

    def __cmp__(self, obj):
        """Compare two _Segments or a _Segment and a _Point.

_Segment comparison works as follows:
1) Get endpoints of both segments S1 and S2
2) Get minima and maxima values for the "x" and "y" coordinates

The return value is determined as follows:

minimum "x" of S1 < minimum "x" of S2 => return -1
maximum "x" of S1 > maximum "x" of S2 => return 1
minimum "y" of S1 < minimum "y" of S2 => return -1
maximum "y" of S1 > maximum "y" of S2 => return 1

Endpoint comparison now is needed to determine if
the _Segments are equal.

if S1.p1.x == S2.p1.x and S1.p1.y == S2.p2.y or
   S1.p1.x == S2.p2.x and s1.p1.y == S2.p2.y ==> return 0

If the the above is not true, then the "x" coordinates of
the first endpoint of each segment are compared.

minimum "x" of S1E1 < minimum "x" of S2E1 => return -1
maximum "x" of S1E1 > maximum "x" of S2E1 => return 1

_Segment/_Point comparison works as follows:
1) Get endpoints of S and coordinates of P
2) Calculate minima and maxima of segment endpoints

The function returns the following

1: P.x < S.xmin
0: S.xmin <= P.x <= S.xmax and S.ymin <= P.y <= S.ymax
-1: Any other condition
        """
        if not isinstance(obj, (point._Point, _Segment)):
            raise TypeError, "Invalid object for Segment comparison: " + `obj`
        _sx1, _sy1 = self.__p1.getCoords()
        _sx2, _sy2 = self.__p2.getCoords()
        _sxmin = min(_sx1, _sx2)
        if isinstance(obj, _Segment):
            _p1, _p2 = obj.getEndpoints()
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            _xmin = min(_x1, _x2)
            # print "sxmin = %g; xmin = %g" % (_sxmin, _xmin)
            if abs(_sxmin - _xmin) < 1e-10:
                _symin = min(_sy1, _sy2)
                _ymin = min(_y1, _y2)
                # print "symin = %g; ymin = %g" % (_symin, _ymin)
                if abs(_symin - _ymin) < 1e-10:
                    _sxmax = max(_sx1, _sx2)
                    _xmax = max(_x1, _x2)
                    # print "sxmax = %g; xmax = %g" % (_sxmax, _xmax)
                    if abs(_sxmax - _xmax) < 1e-10:
                        _symax = max(_sy1, _sy2)
                        _ymax = max(_y1, _y2)
                        # print "symax = %g; ymax = %g" % (_symax, _ymax)
                        if abs(_symax - _ymax) < 1e-10:
                            _x1diff = abs(_sx1 - _x1)
                            _y1diff = abs(_sy1 - _y1)
                            _x2diff = abs(_sx1 - _x2)
                            _y2diff = abs(_sy1 - _y2)
                            if ((_x1diff < 1e-10 and _y1diff < 1e-10) or
                                (_x2diff < 1e-10 and _y2diff < 1e-10)):
                                # print "common endpoints ..."
                                # print "self: " + str(self)
                                # print "obj: " + str(obj)
                                _val = 0
                            elif _sx1 > _x1:
                                _val = 1
                            else:
                                _val = -1
                        elif _symax > _ymax:
                            _val = 1
                        else:
                            _val = -1
                    elif _sxmax > _xmax:
                        _val = 1
                    else:
                        _val = -1
                elif _symin > _ymin:
                    _val = 1
                else:
                    _val = -1
            elif _sxmin > _xmin:
                _val = 1
            else:
                _val = -1
        else:
            _x, _y = obj.getCoords()
            if _x < _sxmin - 1e-10:
                _val = 1
            else:
                _val = -1
                _symin = min(_sy1, _sy2) - 1e-10
                _sxmax = max(_sx1, _sx2) + 1e-10
                _symax = max(_sy1, _sy2) + 1e-10
                if _x < _sxmax and _symin < _y < _symax:
                    _val = 0
        return _val

    def getEndpoints(self):
        """Get the endpoints of the _Segment.

getEndpoints()

This function returns a tuple containing the two _Point objects
that are the endpoints of the segment.
        """
        return (self.__p1, self.__p2)

    def getP1(self):
        """Return the first endpoint _Point of the _Segment.

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint _Point of the _Segment.

setP1(p)
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid Point for p1 endpoint: " + str(p)
        self.__p1 = p

    p1 = property(getP1, setP1, None, "First endpoint of the _Segment.")

    def getP2(self):
        """Return the second endpoint _Point of the _Segment.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the second endpoint _Point of the _Segment.

setP2(p)
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + str(p)
        self.__p2 = p

    p2 = property(getP2, setP2, None, "Second endpoint of the _Segment.")

    def move(self, dx, dy):
        """Move a _Segment.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__p1.move(_dx, _dy)
        self.__p2.move(_dx, _dy)

    def length(self):
        """Return the length of the _Segment.

length()
        """
        return self.__p1 - self.__p2

    def getCoefficients(self):
        """Express the line segment as a function ax + by + c = 0

getCoefficients()

This method returns a tuple of three floats: (a, b, c)
        """
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _a = _y2 - _y1
        _b = _x1 - _x2
        _c = (_x2 * _y1) - (_x1 * _y2)
        return _a, _b, _c

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest _Point on the _Segment to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A _Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by _Point object to an
actual _Point on the _Segment. If the distance between the actual
_Point and the _Point used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def getProjection(self, x, y):
        """Find the projection point of some coordinates on the Segment.

getProjection(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _sqlen = pow((_x2 - _x1), 2) +  pow((_y2 - _y1), 2)
        _rn = ((_x - _x1) * (_x2 - _x1)) + ((_y - _y1) * (_y2 - _y1))
        _r = _rn/_sqlen
        if _r < 0.0 or _r > 1.0:
            return None
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        return _px, _py

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest _Point on the _Segment to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by coordinate pair to an
actual _Point on the _Segment. If the distance between the actual
_Point and the coordinates used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _ap = None
        _pt = util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)
        if _pt is not None:
            _px, _py = _pt
            _ap = point.Point(_px, _py)
        return _ap

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a _Segment exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the _Segment lies within that area. If
the optional argument "fully" is used and is True, then both
endpoints of the _Segment must lie within the boundary.
Otherwise, the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`

        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _pxmin = min(_x1, _x2)
        _pymin = min(_y1, _y2)
        _pxmax = max(_x1, _x2)
        _pymax = max(_y1, _y2)

        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax):
            return True
        if not fully:
            _val = util.in_region(_x1, _y1, _x2, _y2, _xmin, _ymin, _xmax, _ymax)
        return _val

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        """Clip the _Segment using the Liang-Barsky Algorithm.

clipToRegion(xmin, ymin, xmax, ymax)
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        # print "xmin: %g" % _xmin
        # print "ymin: %g" % _ymin
        # print "xmax: %g" % _xmax
        # print "ymax: %g" % _ymax
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        #
        # simple tests to reject line
        #
        if ((_x1 < _xmin and _x2 < _xmin) or
            (_x1 > _xmax and _x2 > _xmax) or
            (_y1 < _ymin and _y2 < _ymin) or
            (_y1 > _ymax and _y2 > _ymax)):
            return None
        #
        # simple tests to accept line
        #
        _coords = None
        if (_xmin < _x1 < _xmax and
            _xmin < _x2 < _xmax and
            _ymin < _y1 < _ymax and
            _ymin < _y2 < _ymax):
            _coords = (_x1, _y1, _x2, _y2)
        else:
            #
            # the Segment can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # The following is the Liang-Barsky Algorithm
            # for segment clipping
            #
            _dx = _x2 - _x1
            _dy = _y2 - _y1
            _P = [-_dx, _dx, -_dy, _dy]
            _q = [(_x1 - _xmin), (_xmax - _x1), (_y1 - _ymin), (_ymax - _y1)]
            _u1 = 0.0
            _u2 = 1.0
            _valid = True
            for _i in range(4):
                _pi = _P[_i]
                _qi = _q[_i]
                if abs(_pi) < 1e-10:
                    if _qi < 0.0:
                        _valid = False
                        break
                else:
                    _r = _qi/_pi
                    if _pi < 0.0:
                        if _r > _u2:
                            _valid = False
                            break
                        if _r > _u1:
                            _u1 = _r
                    else:
                        if _r < _u1:
                            _valid = False
                            break
                        if _r < _u2:
                            _u2 = _r
            if _valid:
                _coords = (((_u1 * _dx) + _x1),
                           ((_u1 * _dy) + _y1),
                           ((_u2 * _dx) + _x1),
                           ((_u2 * _dy) + _y1))
        return _coords


    def clone(self):
        """Create an identical copy of a _Segment.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        return _Segment(_cp1, _cp2)

class old_Segment(_Segment, tree.TreeItem, baseobject.Subpart,
              graphicobject.GraphicObject):
    """The storable and displayed representation of _Segment objects.

The Segment class is derived from _Segment, treeItem, and
GraphicObject classes, so it shares all the attributes and
methods of those classes.
    """

    #
    # The default style for the Segment class
    #

    __defstyle = style.Style(u'Segment Default Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, p1, p2, st=None, lt=None, col=None, t=None):
        """Initialize a Segment.

Segment(p1, p2[, st, lt, col, t])

There are two required arguments:

p1: A Point giving the first Segment endpoint.
p2: A Point giving the second Segment endpoint.

The following arguments are optional:

st: A Style object
lt: A Linetype object that overrides the Style linetype
col: A Color object that overrides the Style color
t: A float that overrides the Style thickness
        """
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 == _p2:
            raise ValueError, "Segments must have two different endpoints."
        _st = st
        if _st is None:
            _st = old_Segment.__defstyle
        _Segment.__init__(self, _p1, _p2)
        tree.TreeItem.__init__(self)
        baseobject.Subpart.__init__(self)
        graphicobject.GraphicObject.__init__(self, _st, lt, col, t)
        _p1.storeUser(self)
        _p2.storeUser(self)

    def isModified(self):
        """Indicate if the Segment has been changed.

isModified()

This method extends the TreeItem::isModified() method.
        """
        _p1, _p2 = self.getEndpoints()
        if _p1.isModified() or _p2.isModified():
            self.modified()
        return tree.TreeItem.isModified(self)

    def setP1(self, p):
        """Set the first endpoint Point of the Segment.

setP1(p)

This method extends _Segment::setP1() and sets the modified
flag indicating that the Segment has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _p1 = self.getP1()
        _Segment.setP1(self, p)
        self.modified()
        _p1.freeUser(self)
        p.storeUser(self)

    p1 = property(_Segment.getP1, setP1, None,
                  "First endpoint of the Segment.")

    def setP2(self, p):
        """Set the second endpoint Point of the Segment.

setP2(p)

This method extends _Segment::setP2() and sets the modified
flag indicating that the Segment has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _p2 = self.getP2()
        _Segment.setP2(self, p)
        self.modified()
        _p2.freeUser(self)
        p.storeUser(self)

    p2 = property(_Segment.getP2, setP2, None,
                  "Second endpoint of the Segment.")

    def move(self, dx, dy):
        """Move a Segment.

move(dx, dy)

This method extends _Segment::move() and sets the modified
flag indicating that the Segment has been altered.
        """
        _Segment.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Segment.

clone()

This method overrides the _Segment::clone() method.
        """
        _p1, _p2 = self.getEndpoints()
        _cp1 = _p1.clone()
        _cp2 = _p2.clone()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Segment(_cp1, _cp2, _st, _lt, _col, _th)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to the point p.

mapPoint(p [, tol])

This function extends _Segment::mapPoint().
        """
        _pt = _Segment.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to a coordinate pair.

mapCoords(x, y[, tol])

This function extends _Segment::mapCoords().
        """
        _pt = _Segment.mapCoords(self, x, y, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

class SegmentTree(maptree.PointMapTree):
    """A Tree class specific for Segment objects.

This class is based on the PointMapTree class, and holds
only Segment objects.
    """
    def __init__(self):
        """Initialize a SegmentTree.

SegmentTree()
        """
        maptree.PointMapTree.__init__(self, Segment)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Segment object held in a SegmentTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: Either a Segment object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Segments are close enough to be considered the same.

If a Segment object is found in the SegmentTree within the
tolerance, the index of that object in the SegmentTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Segment):
            _p1, _p2 = _o.getEndpoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 2:
                raise ValueError, "Tuple must contain exactly two objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
        else:
            raise TypeError, "Invalid type for SegmentTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "SegmentTree in modified state."
        _ro = None
        if len(self):
            _xmin = min(_p1.x, _p2.x)
            _ymin = min(_p1.y, _p2.y)
            _xmax = max(_p1.x, _p2.x)
            _ymax = max(_p1.y, _p2.y)
            # print "xmin: %g" % _xmin
            # print "ymin: %g" % _ymin
            # print "xmax: %g" % _xmax
            # print "ymax: %g" % _ymax
            _scanlist = []
            _lo = _mid = 0
            _hi = len(self)
            # print "lo: 0; hi: %d" % _hi
            while _lo < _hi:
                _mid = (_hi+_lo)//2
                if _mid in _scanlist:
                    break
                _scanlist.append(_mid)
                # print "testing %d" % _mid
                _sp1, _sp2 = self[_mid].getEndpoints()
                # print _sp1
                # print _sp2
                _sxmin = min(_sp1.x, _sp2.x)
                # print "sxmin: %g" % _sxmin
                if (_xmin - _sxmin) > -1e-10:
                    _lo = _mid + 1
                else:
                    _hi = _mid
                    # print "lo: %d; hi: %d" % (_lo, _hi)
            for _i in range(_mid, -1, -1):
                # print "index: %d" % _i
                _seg = self[_i]
                _sp1, _sp2 = _seg.getEndpoints()
                _sxmin = min(_sp1.x, _sp2.x)
                if _sxmin < _xmin - _t - 1e-10:
                    break
                if abs(_sxmin - _xmin) < _t + 1e-10:
                    _symin = min(_sp1.y, _sp2.y)
                    if abs(_symin - _ymin) < _t + 1e-10:
                        _sxmax = max(_sp1.x, _sp2.x)
                        if abs(_sxmax - _xmax) < _t + 1e-10:
                            _symax = max(_sp1.y, _sp2.y)
                            if abs(_symax - _ymax) < _t + 1e-10:
                                _ro = _seg
                                break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a Segment object held in a SegmentTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: Either a Segment object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Segments are close enough to be considered equal.

If a Segment object is found in the SegmentTree within the
tolerance, the index of that object in the SegmentTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Segment):
            _p1, _p2 = _o.getEndpoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 2:
                raise ValueError, "Tuple must contain exactly two objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
        else:
            raise TypeError, "Invalid type for SegmentTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "SegmentTree in modified state."
        _xmin = min(_p1.x, _p2.x)
        _ymin = min(_p1.y, _p2.y)
        _xmax = max(_p1.x, _p2.x)
        _ymax = max(_p1.y, _p2.y)
        _ro = None
        for _sobj in self:
            _sp1, _sp2 = _sobj.getEndpoints()
            _sxmin = min(_sp1.x, _sp2.x)
            if _sxmin > _xmin + _t + 1e-10:
                break
            if abs(_sxmin - _xmin) < _t + 1e-10:
                _symin = min(_sp1.y, _sp2.y)
                if abs(_symin - _ymin) < _t + 1e-10:
                    _sxmax = max(_sp1.x, _sp2.x)
                    if abs(_sxmax - _xmax) < _t + 1e-10:
                        _symax = max(_sp1.y, _sp2.y)
                        if abs(_symax - _ymax) < _t + 1e-10:
                            _ro = _sobj
                            break
        return _ro

#
# Quadtree Segment storage
#

class SegmentQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, Segment):
            raise TypeError, "Invalid Segment object: " + `obj`
        if obj in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _sxmin - 1.0
            _ymin = _symin - 1.0
            _xmax = _sxmax + 1.0
            _ymax = _symax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _sxmin < _xmin:
                _xmin = _sxmin - 1.0
                _resize = True
            if _sxmax > _xmax:
                _xmax = _sxmax + 1.0
                _resize = True
            if _symin < _ymin:
                _ymin = _symin - 1.0
                _resize = True
            if _symax > _ymax:
                _ymax = _symax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, SegmentQuadtree.moveSegment)

    def delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # seg may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _t = tolerance.TOL
        if _alen > 4:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _sdict = {}
        _bailout = False
        _seg = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < (_xmid - _t): # seg on left side
                    _ne = _se = False
                if _sxmin > (_xmid + _t): # seg on right side
                    _nw = _sw = False
                if _symax < (_ymid - _t): # seg below
                    _nw = _ne = False
                if _symin > (_ymid + _t): # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _s in _node.getObjects():
                    _sid = id(_s)
                    if _sid not in _sdict:
                        _p1, _p2 = _s.getEndpoints()
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x1) < _t) and (abs(_py - _y1) < _t)):
                            _px, _py = _p2.getCoords()
                            if ((abs(_px - _x2) < _t) and
                                (abs(_py - _y2) < _t)):
                                _bailout = True
                                _seg = _s
                                break
                        _px, _py = _p2.getCoords()
                        if ((abs(_px - _x1) < _t) and (abs(_py - _y1) < _t)):
                            _px, _py = _p1.getCoords()
                            if ((abs(_px - _x2) < _t) and
                                (abs(_py - _y2) < _t)):
                                _bailout = True
                                _seg = _s
                                break
                        _sdict[_sid] = True
            if _bailout:
                break
        return _seg

    def moveSegment(self, obj, *args):
        if obj not in self:
            raise ValueError, "Segment not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # segment may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _seg = _tsep = None
        _bailout = False
        _sdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _s in _node.getObjects():
                    _sid = id(_s)
                    if _sid not in _sdict:
                        _p1, _p2 = _s.getEndpoints()
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _seg = _s
                            _bailout = True
                            break
                        _px, _py = _p2.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _seg = _s
                            _bailout = True
                            break
                        _sdict[_sid] = True
                    _pt = _s.mapCoords(_x, _y, _t)
                    if _pt is not None:
                        _px, _py = _pt
                        _sep = math.hypot((_px - _x), (_py - _y))
                        if _tsep is None:
                            _tsep = _sep
                            _seg = _s
                        else:
                            if _sep < _tsep:
                                _tsep = _sep
                                _seg = _s
            if _bailout:
                break
        return _seg

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _segs = []
        if not len(self):
            return _segs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _sdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _seg in _node.getObjects():
                    _sid = id(_seg)
                    if _sid not in _sdict:
                        if _seg.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _segs.append(_seg)
                        _sdict[_sid] = True
        return _segs
